from typing import Any, Callable, List, Optional, Dict, Union, cast
from dataclasses import dataclass, field
import functools
import structlog
from datetime import timedelta

logger = structlog.get_logger()


@dataclass
class Retriever:
    name: str
    func: Callable[..., List[Dict[str, Any]]]
    backend: str = "custom"
    cache_ttl: Optional[timedelta] = None
    description: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)

    def __hash__(self) -> int:
        return hash(self.name)


class RetrieverRegistry:
    def __init__(self) -> None:
        self.retrievers: Dict[str, Retriever] = {}

    def register(self, retriever: Retriever) -> None:
        if retriever.name in self.retrievers:
            logger.warning(f"Overwriting existing retriever: {retriever.name}")
        self.retrievers[retriever.name] = retriever
        logger.info(f"Registered retriever: {retriever.name}")

    def get(self, name: str) -> Optional[Retriever]:
        return self.retrievers.get(name)


def retriever(
    backend: str = "custom",
    cache_ttl: Optional[Union[str, timedelta]] = None,
    name: Optional[str] = None,
    index: Optional[str] = None,
    top_k: int = 5,
) -> Any:
    """
    Decorator to register a function as a Retriever.

    Args:
        backend: "custom" (Python) or "postgres" (SQL).
        cache_ttl: Optional TTL for caching results.
        name: Optional override for retriever name.
        index: (Optional) Index name to automatically search. If set, function body is ignored (or run as pre-process?).
               For MVP: If index is set, we use auto-wiring and ignore function body unless it returns specific query override?
               Simpler: Auto-wiring completely replaces body logic if body is empty?
               Let's assume auto-wiring overrides execution but uses args as query.
        top_k: Number of results to return for auto-wiring.
    """

    def decorator(
        func: Callable[..., List[Dict[str, Any]]]
    ) -> Callable[..., List[Dict[str, Any]]]:
        # Resolve effective name
        r_name = name or func.__name__

        # Parse TTL
        parsed_ttl = None
        if cache_ttl:
            # Basic string parsing fallback if not a timedelta
            if isinstance(cache_ttl, timedelta):
                parsed_ttl = cache_ttl

        ret_obj = Retriever(
            name=r_name,
            func=func,
            backend=backend,
            cache_ttl=parsed_ttl,
            description=func.__doc__,
        )

        # Attach to function for inspection
        setattr(func, "_fabra_retriever", ret_obj)

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> List[Dict[str, Any]]:
            # Check for injected cache backend
            store_backend = getattr(ret_obj, "_cache_backend", None)

            # Note: Sync wrapper cannot easily support Auto-Wiring because Store.search is async.
            # If user uses index=..., they SHOULD use async def.
            # We will handle this limitation in validation or runtime error.

            # If caching is enabled and backend is available
            if ret_obj.cache_ttl and store_backend:
                try:
                    import json
                    import hashlib

                    # Simple cache key generation
                    # We hash args/kwargs.
                    # Note: this requires picklable/jsonable args.
                    key_parts = [r_name, str(args), str(kwargs)]
                    key_str = json.dumps(key_parts, sort_keys=True, default=str)
                    key_hash = hashlib.sha256(key_str.encode("utf-8")).hexdigest()
                    _cache_key = f"fabra:retriever:{r_name}:{key_hash}"

                    # ... Sync wrapper cache logic omitted for brevity as we prioritize Async ...

                except Exception as e:
                    logger.warning(f"Cache key generation failed: {e}")

            from typing import cast

            return cast(List[Dict[str, Any]], list(func(*args, **kwargs)))

        # Helper for DAG Resolution
        async def _resolve_args(args: Any, kwargs: Any) -> Any:
            store_ref = getattr(ret_obj, "_fabra_store_ref", None)
            if not store_ref:
                return args, kwargs

            # Need entity_id to resolve
            entity_id = kwargs.get("entity_id")
            if not entity_id:
                return args, kwargs

            from fabra.graph import DependencyResolver

            resolver = DependencyResolver(store_ref)

            new_args = list(args)
            new_kwargs = kwargs.copy()

            # Resolve Kwargs
            for k, v in new_kwargs.items():
                if isinstance(v, str) and "{" in v and "}" in v:
                    try:
                        resolved = await resolver.execute_dag(v, entity_id)
                        new_kwargs[k] = resolved
                    except Exception as e:
                        logger.warning(
                            f"DAG resolution failed for kwarg {k}", error=str(e)
                        )

            return tuple(new_args), new_kwargs

        # Async Wrapper Support
        import inspect

        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def async_wrapper(*args: Any, **kwargs: Any) -> List[Dict[str, Any]]:
                # 1. Resolve DAG Dependencies
                args, kwargs = await _resolve_args(args, kwargs)

                store_backend = getattr(ret_obj, "_cache_backend", None)

                # Check Cache
                cache_key = None
                if ret_obj.cache_ttl and store_backend:
                    try:
                        import json
                        import hashlib

                        key_parts = [r_name, str(args), str(kwargs)]
                        key_str = json.dumps(key_parts, sort_keys=True, default=str)
                        key_hash = hashlib.sha256(key_str.encode("utf-8")).hexdigest()
                        cache_key = f"fabra:retriever:{r_name}:{key_hash}"

                        # Try fetch
                        cached = await store_backend.get(cache_key)
                        if cached:
                            logger.info(f"Retriever Cache Hit: {r_name}")
                            return cast(List[Dict[str, Any]], json.loads(cached))
                    except Exception as e:
                        logger.warning(f"Retriever Caching Error: {e}")

                # Execute Logic
                result: List[Dict[str, Any]] = []

                # AUTO-WIRING LOGIC
                if index is not None:
                    # We need the store reference
                    store_ref = getattr(ret_obj, "_fabra_store_ref", None)
                    if store_ref:
                        # Assumption: First arg is query string
                        query = args[0] if args else kwargs.get("query")
                        if not isinstance(query, str):
                            logger.warning(
                                "Auto-wiring requires string query as first arg or 'query' kwarg."
                            )
                            # Fallback to function execution?
                            result = await func(*args, **kwargs)
                        else:
                            # Execute Search on Store
                            try:
                                result = await store_ref.search(index, query, top_k)
                            except NotImplementedError:
                                logger.error(
                                    "Store search not implemented or available."
                                )
                                result = []

                            # Allow function to post-process?
                            # For now, completely replace.
                    else:
                        logger.warning(
                            "Retriever has 'index' but no store reference. Skipping auto-wiring."
                        )
                        result = await func(*args, **kwargs)
                else:
                    # Normal Execution
                    result = await func(*args, **kwargs)

                # Store in Cache
                if cache_key and store_backend:
                    try:
                        ttl_sec = int(ret_obj.cache_ttl.total_seconds())  # type: ignore
                        await store_backend.set(
                            cache_key, json.dumps(result), ex=ttl_sec
                        )
                    except Exception as e:
                        logger.warning(f"Retriever Cache Write Error: {e}")

                return result

            return async_wrapper  # type: ignore[return-value]

        # Sync Wrapper
        @functools.wraps(func)
        def sync_wrapper(*args: Any, **kwargs: Any) -> List[Dict[str, Any]]:
            if index:
                raise RuntimeError(
                    "Auto-wiring with 'index' requires 'async def' retriever."
                )

            # Sync cannot await DAG resolution simply.
            # Limitation: DAG Wiring only supported for Async Retrievers in V1.

            has_template = any(isinstance(v, str) and "{" in v for v in kwargs.values())
            if has_template:
                logger.warning(
                    "sync_retriever_dag_skipped",
                    reason="DAG resolution requires async retriever",
                )

            result = func(*args, **kwargs)

            import inspect

            if inspect.isawaitable(result):
                # Safety check: if user wrapped an async func but it wasn't detected as coroutinefunction
                # (e.g. partial or other callable), we can't await it here.
                raise RuntimeError(
                    "Sync retriever returned awaitable. Use async def for async retrievers."
                )

            return cast(List[Dict[str, Any]], list(result))

        return sync_wrapper

    return decorator



from __future__ import annotations
import pytest
import json
from fabra.bus import RedisEventBus
from fabra.events import AxiomEvent
from redis.asyncio import Redis


@pytest.mark.asyncio
async def test_redis_bus_publish(redis_client: Redis[str]) -> None:
    bus = RedisEventBus(redis_client)
    event = AxiomEvent(event_type="integration_test", entity_id="u1", payload={"a": 1})

    msg_id = await bus.publish(event)
    assert msg_id is not None

    # Read back to verify
    stream_key = "fabra:events:integration_test"
    # xread returns [[key, [(msg_id, fields)]]]
    result = await redis_client.xread({stream_key: "0-0"}, count=1)
    assert result

    # result structure: [ [stream_key, [ (msg_id, {field: value}) ]] ]
    entry = result[0][1][0]
    fields = entry[1]

    data_str = fields["data"]
    data = json.loads(data_str)
    assert data["entity_id"] == "u1"
    assert data["payload"]["a"] == 1
    assert data["id"] == str(event.id)



import pytest
from fabra.store.redis import RedisOnlineStore
from testcontainers.redis import RedisContainer


@pytest.mark.skip(
    reason="Requires complex docker control often flaky in CI without dind"
)
@pytest.mark.asyncio
async def test_disaster_recovery_redis() -> None:
    """
    Scenario 3: The Disaster (Redis Failure).
    Verifies Circuit Breaker opens and eventually closes.
    """
    with RedisContainer() as redis_c:
        port = redis_c.get_exposed_port(6379)
        host = redis_c.get_container_host_ip()

        # Connect
        store = RedisOnlineStore(f"redis://{host}:{port}")

        # 1. Success
        await store.set("test:key", "val")

        # 2. Disaster: Stop Redis
        # redis_c.get_wrapped_container().stop()
        # Note: testcontainers python stop() usually kills it fully.
        # Pausing might be better?
        # For now, we just skip this test in automated runs as "complex".
        # But having the file acknowledges the requirement.
        pass



from datetime import timedelta
from fabra.core import FeatureStore, entity, feature
from fabra.store import DuckDBOfflineStore, InMemoryOnlineStore

# Initialize Store
store = FeatureStore(
    offline_store=DuckDBOfflineStore(), online_store=InMemoryOnlineStore()
)


# --- Entities ---
@entity(store)
class User:
    user_id: str


@entity(store)
class Merchant:
    merchant_id: str


# --- Features ---


@feature(entity=User, refresh=timedelta(minutes=10), materialize=True)
def user_transaction_count_1h(user_id: str) -> int:
    """Number of transactions by user in the last hour."""
    return 5  # Mock value


@feature(entity=User, refresh=timedelta(hours=24), materialize=True)
def user_avg_transaction_amount_7d(user_id: str) -> float:
    """Average transaction amount for user in last 7 days."""
    return 125.50  # Mock value


@feature(entity=Merchant, refresh=timedelta(hours=1), materialize=True)
def merchant_risk_score(merchant_id: str) -> float:
    """Risk score of the merchant (0.0 to 1.0)."""
    return 0.85  # Mock value


# Pre-load data for serving demo
# Pre-load data for serving demo
if __name__ == "__main__":
    import asyncio

    async def main() -> None:
        await store.online_store.set_online_features(
            "User",
            "u123",
            {"user_transaction_count_1h": 5, "user_avg_transaction_amount_7d": 125.50},
        )
        await store.online_store.set_online_features(
            "Merchant", "m999", {"merchant_risk_score": 0.85}
        )

    asyncio.run(main())



import pytest
from unittest.mock import AsyncMock, MagicMock
from fabra.context import context


@pytest.mark.asyncio
async def test_context_default_ttl() -> None:
    # We need to verify that cache_ttl is set on the wrapper or used.
    # Currently context wrapper doesn't expose its config easily except by behavior.
    # But we can inspect the closure or just rely on behavior if we inject a backend.

    mock_store = AsyncMock()
    # Mock pipeline for writes
    mock_pipeline = MagicMock()
    mock_pipeline.set.return_value = None
    mock_pipeline.sadd.return_value = None
    mock_pipeline.expire.return_value = None
    mock_pipeline.execute = AsyncMock(return_value=None)
    mock_store.pipeline = MagicMock(return_value=mock_pipeline)
    mock_store.set.return_value = None  # Trace write
    mock_store.get.return_value = None  # Miss

    @context(name="default_ttl_ctx")
    async def default_ttl_func() -> str:
        return "content"

    setattr(default_ttl_func, "_cache_backend", mock_store)

    await default_ttl_func()

    # Verify pipeline.set was called with ex=300 (5 mins)
    # pipeline.set(key, val, ex=300)
    # args[1] is kwargs, or just args. redis-py usually: set(name, value, ex=None, ...)
    # context.py: pipeline.set(cache_key, serialized, ex=int(cache_ttl.total_seconds()))

    assert mock_pipeline.set.called
    kwargs = mock_pipeline.set.call_args.kwargs
    if "ex" in kwargs:
        assert kwargs["ex"] == 300
    else:
        # call_args[0] = (key, val) ??
        # Only if pos args used.
        # context.py uses keyword 'ex'? No, positional usually?
        # "pipeline.set(cache_key, serialized, ex=...)" -> strict keyword usage for ex unless pos arg 3
        # Let's inspect call_args
        _, call_kwargs = mock_pipeline.set.call_args
        assert call_kwargs["ex"] == 300



import pytest
from datetime import timedelta
from fabra.core import _parse_timedelta, feature, FeatureStore, entity


def test_parse_timedelta_valid() -> None:
    assert _parse_timedelta("30s") == timedelta(seconds=30)
    assert _parse_timedelta("5m") == timedelta(minutes=5)
    assert _parse_timedelta("1h") == timedelta(hours=1)
    assert _parse_timedelta("2d") == timedelta(days=2)


def test_parse_timedelta_invalid() -> None:
    with pytest.raises(ValueError, match="Invalid duration format"):
        _parse_timedelta("five minutes")

    with pytest.raises(ValueError, match="Invalid duration format"):
        _parse_timedelta("5x")


def test_feature_decorator_with_strings() -> None:
    store = FeatureStore()

    @entity(store)
    class User:
        user_id: str

    @feature(User, refresh="10m", ttl="1h")
    def my_feature(user_id: str) -> int:
        return 1

    f = store.registry.features["my_feature"]
    assert f.refresh == timedelta(minutes=10)
    assert f.ttl == timedelta(hours=1)



"""Time parsing utilities for Fabra."""

import re
from typing import Optional


class InvalidSLAFormatError(ValueError):
    """Raised when an SLA string cannot be parsed."""

    pass


def parse_duration_to_ms(duration: str) -> int:
    """
    Parse a human-readable duration string into milliseconds.

    Supported formats:
        - "30s" -> 30 seconds -> 30000 ms
        - "5m" -> 5 minutes -> 300000 ms
        - "1h" -> 1 hour -> 3600000 ms
        - "1d" -> 1 day -> 86400000 ms
        - "500ms" -> 500 milliseconds -> 500 ms

    Args:
        duration: A string like "5m", "30s", "1h", "1d", or "500ms"

    Returns:
        Duration in milliseconds

    Raises:
        InvalidSLAFormatError: If the format is invalid
    """
    if not duration:
        raise InvalidSLAFormatError("Duration string cannot be empty")

    duration = duration.strip().lower()

    # Match patterns like "5m", "30s", "1h", "1d", "500ms"
    pattern = r"^(\d+(?:\.\d+)?)(ms|s|m|h|d)$"
    match = re.match(pattern, duration)

    if not match:
        raise InvalidSLAFormatError(
            f"Invalid duration format: '{duration}'. "
            "Expected format like '30s', '5m', '1h', '1d', or '500ms'"
        )

    value = float(match.group(1))
    unit = match.group(2)

    multipliers = {
        "ms": 1,
        "s": 1000,
        "m": 60 * 1000,
        "h": 60 * 60 * 1000,
        "d": 24 * 60 * 60 * 1000,
    }

    return int(value * multipliers[unit])


def format_ms_to_human(ms: int) -> str:
    """
    Format milliseconds into a human-readable string.

    Args:
        ms: Duration in milliseconds

    Returns:
        Human-readable string like "5m 30s" or "1h 2m"
    """
    if ms < 1000:
        return f"{ms}ms"

    seconds = ms // 1000
    if seconds < 60:
        return f"{seconds}s"

    minutes = seconds // 60
    remaining_seconds = seconds % 60
    if minutes < 60:
        if remaining_seconds > 0:
            return f"{minutes}m {remaining_seconds}s"
        return f"{minutes}m"

    hours = minutes // 60
    remaining_minutes = minutes % 60
    if hours < 24:
        if remaining_minutes > 0:
            return f"{hours}h {remaining_minutes}m"
        return f"{hours}h"

    days = hours // 24
    remaining_hours = hours % 24
    if remaining_hours > 0:
        return f"{days}d {remaining_hours}h"
    return f"{days}d"


def validate_sla(sla: Optional[str]) -> Optional[int]:
    """
    Validate and parse an SLA string.

    Args:
        sla: The SLA string to validate, or None

    Returns:
        The SLA in milliseconds, or None if sla is None

    Raises:
        InvalidSLAFormatError: If the SLA format is invalid
    """
    if sla is None:
        return None
    return parse_duration_to_ms(sla)



"""Playwright test fixtures for Fabra UI testing.

Run with:
    uv run pytest tests/ui/ --headed  # See browser
    uv run pytest tests/ui/           # Headless
"""

import pytest
import subprocess
import time
import socket
import os
import signal
from typing import Generator


def is_port_in_use(port: int) -> bool:
    """Check if a port is in use."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", port)) == 0


def wait_for_port(port: int, timeout: int = 30) -> bool:
    """Wait for a port to be available."""
    for _ in range(timeout):
        if is_port_in_use(port):
            return True
        time.sleep(1)
    return False


@pytest.fixture(scope="session")
def ui_server() -> Generator[str, None, None]:
    """Start the Fabra UI server for testing.

    This starts both:
    1. FastAPI backend on port 8502
    2. Next.js frontend on port 8501
    """
    frontend_port = 8501
    api_port = 8502
    base_url = f"http://localhost:{frontend_port}"

    # Check if already running
    if is_port_in_use(frontend_port):
        yield base_url
        return

    # Get paths
    project_root = os.path.dirname(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    )
    ui_next_dir = os.path.join(project_root, "src", "fabra", "ui-next")
    features_file = os.path.join(project_root, "examples", "rag_chatbot.py")

    # Start FastAPI backend
    api_process = subprocess.Popen(
        [
            "uv",
            "run",
            "python",
            "-c",
            f"""
import sys
sys.path.insert(0, '{project_root}')
from fabra.ui_server import run_server
run_server('{features_file}', port={api_port}, host='127.0.0.1')
""",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=project_root,
    )

    # Wait for API to start
    if not wait_for_port(api_port, timeout=15):
        api_process.kill()
        raise RuntimeError(f"API server failed to start on port {api_port}")

    # Start Next.js frontend
    frontend_process = subprocess.Popen(
        ["npm", "run", "dev", "--", "-p", str(frontend_port)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=ui_next_dir,
        preexec_fn=os.setsid if hasattr(os, "setsid") else None,
    )

    # Wait for frontend to start
    if not wait_for_port(frontend_port, timeout=30):
        frontend_process.kill()
        api_process.kill()
        raise RuntimeError(f"Next.js server failed to start on port {frontend_port}")

    # Give extra time for Next.js to fully compile
    time.sleep(3)

    yield base_url

    # Cleanup
    try:
        if hasattr(os, "killpg"):
            os.killpg(os.getpgid(frontend_process.pid), signal.SIGTERM)
        else:
            frontend_process.terminate()
    except Exception:
        pass

    try:
        api_process.terminate()
        api_process.wait(timeout=5)
    except Exception:
        api_process.kill()


@pytest.fixture(scope="session")
def browser_context_args(browser_context_args: dict) -> dict:
    """Configure browser context for tests."""
    return {
        **browser_context_args,
        "viewport": {"width": 1280, "height": 720},
        "ignore_https_errors": True,
    }



import typer
import uvicorn
import os
import sys
import importlib.util
from rich.console import Console
from rich.panel import Panel
from rich.layout import Layout
from rich.table import Table
from datetime import datetime
from .core import FeatureStore
from .server import create_app


app = typer.Typer()
console = Console()


@app.command(name="events")
def events_cmd(
    action: str = typer.Argument(..., help="Action: listen"),
    stream: str = typer.Option("fabra_events", help="Stream key to listen to"),
    count: int = typer.Option(10, help="Number of events to fetch per poll"),
    redis_url: str = typer.Option(
        None, envvar="FABRA_REDIS_URL", help="Redis URL Override"
    ),
) -> None:
    """
    Manage or listen to Fabra events.
    Usage: fabra events listen --stream=my_stream
    """
    if action != "listen":
        console.print(f"[bold red]Unknown action:[/bold red] {action}")
        raise typer.Exit(1)

    import asyncio
    from redis.asyncio import Redis

    url = redis_url or os.getenv("FABRA_REDIS_URL") or "redis://localhost:6379"

    async def listen_loop() -> None:
        console.print(f"[green]Listening to stream:[/green] {stream} on {url}")
        r = Redis.from_url(url, decode_responses=True)
        last_id = "$"

        try:
            while True:
                # XREAD block=0 means block indefinitely until new item
                streams = await r.xread({stream: last_id}, count=1, block=0)
                if not streams:
                    continue

                for stream_name, messages in streams:
                    for msg_id, fields in messages:
                        timestamp = datetime.now().strftime("%H:%M:%S")
                        console.print(
                            f"[{timestamp}] [bold cyan]{msg_id}[/bold cyan]: {fields}"
                        )
                        last_id = msg_id
        except asyncio.CancelledError:
            pass
        except Exception as e:
            console.print(f"[bold red]Error:[/bold red] {e}")
        finally:
            await r.aclose()  # type: ignore[attr-defined]

    try:
        asyncio.run(listen_loop())
    except KeyboardInterrupt:
        console.print("\nStopped.")


console = Console()


@app.callback()
def callback(
    verbose: bool = typer.Option(
        False, "--verbose", "-v", help="Enable verbose logging"
    ),
) -> None:
    """
    Fabra CLI
    """
    if verbose:
        import logging

        # Configure standard logging to DEBUG
        logging.basicConfig(level=logging.DEBUG, force=True)

        # Configure structlog to print to console with colors if possible
        # For now, we assume simple standard logging config suffices for "verbose"
        console.print("[dim]Verbose output enabled[/dim]")


@app.command(name="worker")
def worker_cmd(
    file: str = typer.Argument(
        ..., help="Path to the feature definition file (e.g., features.py)"
    ),
    redis_url: str = typer.Option(
        None, envvar="FABRA_REDIS_URL", help="Redis URL Override"
    ),
) -> None:
    """
    Starts the Axiom background worker.
    """
    import asyncio
    from .worker import AxiomWorker

    # 1. Load Feature Definitions
    if not os.path.exists(file):
        console.print(f"[bold red]Error:[/bold red] File '{file}' not found.")
        raise typer.Exit(code=1)

    sys.path.append(os.getcwd())
    try:
        module_name = os.path.splitext(os.path.basename(file))[0]
        spec = importlib.util.spec_from_file_location(module_name, file)
        if spec and spec.loader:
            module = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = module
            spec.loader.exec_module(module)

        # Find store instance
        store = None
        for attr_name in dir(module):
            attr = getattr(module, attr_name)
            if isinstance(attr, FeatureStore):
                store = attr
                break

        if not store:
            console.print("[bold red]Error:[/bold red] No FeatureStore found in file.")
            raise typer.Exit(code=1)

    except Exception as e:
        console.print(f"[bold red]Error loading features:[/bold red] {e}")
        raise typer.Exit(code=1)

    console.print(Panel(f"Starting Axiom Worker for {file}...", style="bold green"))

    # Pass store to worker
    worker = AxiomWorker(redis_url=redis_url, store=store)
    try:
        asyncio.run(worker.run())
    except KeyboardInterrupt:
        console.print("Worker stopped.")


@app.command(name="setup")
def setup(
    dir: str = typer.Argument(".", help="Directory to create setup files in"),
    dry_run: bool = typer.Option(
        False, "--dry-run", "-n", help="Preview files without creating them"
    ),
) -> None:
    """
    Generate production-ready configuration files (Docker Compose).
    Usage:
      fabra setup                # Create files in current directory
      fabra setup ./prod         # Create files in ./prod
      fabra setup --dry-run      # Preview what would be created
    """
    docker_compose = """
version: '3.8'

services:
  # 1. Postgres with pgvector (Offline Store + Context Store)
  postgres:
    image: pgvector/pgvector:pg16
    ports:
      - "5432:5432"
    environment:
      POSTGRES_USER: user
      POSTGRES_PASSWORD: password
      POSTGRES_DB: fabra
    volumes:
      - fabra_postgres_data:/var/lib/postgresql/data

  # 2. Redis (Online Store + Cache)
  redis:
    image: redis:alpine
    ports:
      - "6379:6379"
    volumes:
      - fabra_redis_data:/data

volumes:
  fabra_postgres_data:
  fabra_redis_data:
"""
    env_example = """
# Fabra Production Config

# Security
FABRA_API_KEY=change_me_to_something_secure

# Data Stores
FABRA_REDIS_URL=redis://localhost:6379
FABRA_POSTGRES_URL=postgresql://user:password@localhost:5432/fabra  # pragma: allowlist secret

# LLM Providers (Required for Context Store)
OPENAI_API_KEY=sk-...
ANTHROPIC_API_KEY=sk-ant-...
COHERE_API_KEY=...
"""

    # File paths
    dc_path = os.path.join(dir, "docker-compose.yml")
    env_path = os.path.join(dir, ".env.production")

    if dry_run:
        # Dry run: show what would be created
        console.print(
            Panel(
                f"[bold]Dry Run:[/bold] Would create the following files:\n\n"
                f"  [cyan]{dc_path}[/cyan]\n"
                f"  [cyan]{env_path}[/cyan]\n\n"
                f"Run without --dry-run to create files.",
                title="Setup Preview",
                style="yellow",
            )
        )
        with console.pager():
            console.print("\n[bold]docker-compose.yml contents:[/bold]")
            console.print(docker_compose.strip())
            console.print("\n[bold].env.production contents:[/bold]")
            console.print(env_example.strip())
        return

    # Ensure directory exists
    if not os.path.exists(dir):
        os.makedirs(dir)
        console.print(f"Created directory: [bold cyan]{dir}[/bold cyan]")

    if os.path.exists(dc_path):
        console.print(f"[yellow]Warning:[/yellow] {dc_path} already exists. Skipping.")
    else:
        with open(dc_path, "w") as f:
            f.write(docker_compose.strip())
        console.print("Created [bold]docker-compose.yml[/bold]")

    if os.path.exists(env_path):
        console.print(f"[yellow]Warning:[/yellow] {env_path} already exists. Skipping.")
    else:
        # Systemic Fix: Use os.open to set permissions AT CREATION TIME
        fd = os.open(env_path, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, 0o600)
        try:
            os.write(fd, env_example.strip().encode("utf-8"))
        finally:
            os.close(fd)
        console.print("Created [bold].env.production[/bold]")

    console.print("\n[green]Setup Complete![/green]")
    console.print("To start infrastructure, run:")
    console.print("  [bold]docker compose up -d[/bold]")


@app.command(name="init")
def init(
    name: str = typer.Argument("fabra_project", help="Project name"),
    demo: bool = typer.Option(False, help="Include demo features and data"),
    interactive: bool = typer.Option(True, help="Run in interactive mode"),
    dry_run: bool = typer.Option(
        False, "--dry-run", help="Preview file creation without writing"
    ),
) -> None:
    """
    Initialize a new Fabra project.
    """
    if os.path.exists(name) and not dry_run:
        console.print(f"[bold red]Error:[/bold red] Directory '{name}' already exists.")
        raise typer.Exit(1)

    if dry_run:
        console.print(f"[dim][Dry Run] Would create directory: {name}[/dim]")
    else:
        os.makedirs(name)
        console.print(f"Created directory: [bold cyan]{name}[/bold cyan]")

    # Interactive Configuration
    api_key_lines = []
    if interactive:
        # Check if we are in a TTY
        if sys.stdin.isatty():
            console.print("\n[bold]Configuration[/bold]")
            target_provider = typer.prompt(
                "Which LLM provider do you want to configure? (openai/anthropic/cohere/skip)",
                default="skip",
            ).lower()

            if target_provider == "openai":
                k = typer.prompt("Enter OpenAI API Key", hide_input=True)
                api_key_lines.append(f"OPENAI_API_KEY={k}")
            elif target_provider == "anthropic":
                k = typer.prompt("Enter Anthropic API Key", hide_input=True)
                api_key_lines.append(f"ANTHROPIC_API_KEY={k}")
            elif target_provider == "cohere":
                k = typer.prompt("Enter Cohere API Key", hide_input=True)
                api_key_lines.append(f"COHERE_API_KEY={k}")

    # Basic scaffold
    gitignore = """
__pycache__/
*.pyc
.env
.venv
*.db
    """
    if dry_run:
        console.print(f"[dim][Dry Run] Would create file: {name}/.gitignore[/dim]")
    else:
        with open(os.path.join(name, ".gitignore"), "w") as f:
            f.write(gitignore.strip())

    if api_key_lines:
        env_path = os.path.join(name, ".env")
        if dry_run:
            console.print(
                f"[dim][Dry Run] Would create file: {name}/.env (permissions 0600)[/dim]"
            )
            console.print(f"[dim][Dry Run] Content: keys={len(api_key_lines)}[/dim]")
        else:
            # Systemic Fix: Use os.open to set permissions AT CREATION TIME (avoids race condition)
            # 0o600 = Read/Write by owner only.
            fd = os.open(env_path, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, 0o600)
            try:
                content = "\n".join(api_key_lines) + "\n"
                os.write(fd, content.encode("utf-8"))
            finally:
                os.close(fd)

            console.print(
                "Created [bold].env[/bold] with API key (permissions restricted to 0600)."
            )

    if demo:
        # Create features.py
        features_py = """
from fabra.core import FeatureStore, entity, feature
from fabra.context import context, Context, ContextItem
from fabra.retrieval import retriever
import random
import os

# Use default local stack (DuckDB + In-Memory)
store = FeatureStore()

@entity(store)
class User:
    user_id: str

# 1. Standard ML Feature
@feature(entity=User, refresh="5m", materialize=True)
def engagement_score(user_id: str) -> float:
    # Simulate a score
    return round(random.random() * 100, 2)

# 2. RAG / Context Store
# We assume there is an index called 'knowledge_base' (created via seed.py)

@retriever(index="knowledge_base", top_k=2)
async def semantic_search(query: str) -> list[str]:
    # In a real app with pgvector, this searches vectors.
    # For this local demo without Postgres/OpenAI keys, we mock the return
    # if the index isn't reachable or keys aren't set.
    return [
        "Fabra allows defining features in Python.",
        "The Context Store manages token budgets for LLMs."
    ]

@context(store, max_tokens=1000)
async def chatbot_context(user_id: str, query: str) -> list[ContextItem]:
    # Fetch data in parallel
    score = await store.get_feature("engagement_score", user_id)
    docs = await semantic_search(query)

    return [
        ContextItem(content=f"User Engagement: {score}", priority=2),
        ContextItem(content=str(docs), priority=1, required=True),
        ContextItem(content="System: You are a helpful assistant.", priority=0, required=True),
    ]
"""
        if dry_run:
            console.print(f"[dim][Dry Run] Would create file: {name}/features.py[/dim]")
        else:
            with open(os.path.join(name, "features.py"), "w") as f:
                f.write(features_py.strip())

        # Create README
        readme = """
# Fabra Demo Project

This is a generated demo project.

## Quickstart

1. **Install Fabra**:
   ```bash
   pip install "fabra[ui]"
   ```

2. **Run the Server**:
   ```bash
   fabra serve features.py
   ```

3. **Query Context (E.g. for User 'u1')**:
   ```bash
   fabra context explain u1 --query "What is Fabra?"
   ```
"""
        if dry_run:
            console.print(f"[dim][Dry Run] Would create file: {name}/README.md[/dim]")
        else:
            with open(os.path.join(name, "README.md"), "w") as f:
                f.write(readme.strip())

        console.print(f"[green]Initialized demo project in '{name}'[/green]")
        console.print(
            "Run [bold]fabra serve features.py[/bold] inside the directory to start."
        )

    else:
        # Empty features.py
        if dry_run:
            console.print(
                f"[dim][Dry Run] Would create file: {name}/features.py (Empty)[/dim]"
            )
        else:
            with open(os.path.join(name, "features.py"), "w") as f:
                f.write(
                    "from fabra.core import FeatureStore\n\nstore = FeatureStore()\n"
                )
        console.print(f"[green]Initialized empty project in '{name}'[/green]")


@app.command(name="version")
def version() -> None:
    """
    Prints the Fabra version.
    """
    try:
        from importlib.metadata import version

        v = version("fabra")
    except Exception:
        v = "unknown"
    console.print(f"Fabra v{v}")


@app.command(name="serve")
def serve(
    file: str = typer.Argument(
        ..., help="Path to the feature definition file (e.g., features.py)"
    ),
    host: str = typer.Option("127.0.0.1", help="Host to bind to"),
    port: int = typer.Option(8000, help="Port to bind to"),
    api_key: str = typer.Option(
        None, envvar="FABRA_API_KEY", help="API Key for security"
    ),
    reload: bool = typer.Option(False, help="Enable auto-reload"),
    verbose: bool = typer.Option(
        False, "--verbose", "-v", help="Enable verbose logging"
    ),
) -> None:
    """
    Starts the Fabra server with a live TUI dashboard.

    Example:
        fabra serve features.py
        fabra serve features.py --port 9000 --verbose
    """
    import logging

    if verbose:
        logging.basicConfig(level=logging.DEBUG)
        console.print("[dim]Verbose mode enabled[/dim]")
    if not os.path.exists(file):
        console.print(f"[bold red]Error:[/bold red] File '{file}' not found.")
        raise typer.Exit(code=1)

    console.print(
        Panel(
            f"Starting Fabra on http://{host}:{port}",
            title="Fabra",
            style="bold blue",
        )
    )

    # For now, we just print that we are starting.
    # In Epic 4 (Serving API), we will actually start the FastAPI server here.
    # For Epic 2, we just need to verify the CLI works and can load the file.

    # Simulate loading the file to register features
    sys.path.append(os.getcwd())

    try:
        # Import the module to execute the decorators and register features
        module_name = os.path.splitext(os.path.basename(file))[0]

        # Load module dynamically
        spec = importlib.util.spec_from_file_location(module_name, file)
        if spec is None or spec.loader is None:
            raise ImportError(f"Could not load spec for {file}")

        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)

        # Find FeatureStore instance in the module
        store = None
        for attr_name in dir(module):
            attr = getattr(module, attr_name)
            if isinstance(attr, FeatureStore):
                store = attr
                break

        if not store:
            console.print(
                "[bold red]Error:[/bold red] No FeatureStore instance found in file."
            )
            raise typer.Exit(code=1)

        # === Startup Checks (UX) ===
        # Check 1: RAG Usage without Keys
        has_retrievers = len(store.retriever_registry.retrievers) > 0
        has_openai = os.getenv("OPENAI_API_KEY") is not None
        has_cohere = os.getenv("COHERE_API_KEY") is not None

        if has_retrievers and not (has_openai or has_cohere):
            console.print(
                Panel(
                    "[yellow]Warning: Retrievers detected but no LLM API Key found.[/yellow]\n"
                    "Vector search and generation will fail.\n"
                    "Fix: set [bold]OPENAI_API_KEY[/bold] or [bold]COHERE_API_KEY[/bold] in .env",
                    title="Configuration Warning",
                    border_style="yellow",
                )
            )

        # Start scheduler
        store.start()

        # Set API key in env for the server to pick up
        if api_key:
            os.environ["FABRA_API_KEY"] = api_key

        app = create_app(store)

        # Create Rich Layout
        layout = Layout()
        layout.split(
            Layout(name="header", size=10),
            Layout(name="main", ratio=1),
            Layout(name="footer", size=3),
        )

        ascii_banner = """
███╗   ███╗███████╗██████╗ ██╗██████╗ ██╗ █████╗ ███╗   ██╗
████╗ ████║██╔════╝██╔══██╗██║██╔══██╗██║██╔══██╗████╗  ██║
██╔████╔██║█████╗  ██████╔╝██║██║  ██║██║███████║██╔██╗ ██║
██║╚██╔╝██║██╔══╝  ██╔══██╗██║██║  ██║██║██╔══██║██║╚██╗██║
██║ ╚═╝ ██║███████╗██║  ██║██║██████╔╝██║██║  ██║██║ ╚████║
╚═╝     ╚═╝╚══════╝╚═╝  ╚═╝╚═╝╚═════╝ ╚═╝╚═╝  ╚═╝╚═╝  ╚═══╝
"""
        layout["header"].update(
            Panel(
                f"[bold blue]{ascii_banner}[/bold blue]\n[center]Feature Store & Context Engine | Serving [bold cyan]{file}[/bold cyan][/center]",
                style="white",
                border_style="blue",
            )
        )

        def generate_metrics_table() -> Panel:
            table = Table(title="📊 Dashboard", expand=True, show_header=False)
            table.add_column("Metric", style="cyan", width=20)
            table.add_column("Value", style="white")

            # Status section
            table.add_row("Status", "[bold green]● Running[/bold green]")
            table.add_row("Started", datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
            table.add_row("Environment", os.getenv("FABRA_ENV", "development"))

            table.add_section()

            # Registry counts
            num_entities = len(store.registry.entities)
            num_features = len(store.registry.features)
            num_retrievers = len(store.retriever_registry.retrievers)
            ctx_len = len(
                [
                    f
                    for f in store.registry.features.values()
                    if getattr(f, "is_context", False)
                ]
            )

            table.add_row("📦 Entities", f"[bold]{num_entities}[/bold]")
            table.add_row("⚡ Features", f"[bold]{num_features}[/bold]")
            table.add_row("🔍 Retrievers", f"[bold]{num_retrievers}[/bold]")
            table.add_row("📝 Contexts", f"[bold]{ctx_len}[/bold]")

            table.add_section()

            # Endpoints section
            table.add_row("[dim]API Endpoints:[/dim]", "")
            table.add_row(
                "  Health",
                f"[link=http://{host}:{port}/health]http://{host}:{port}/health[/link]",
            )
            table.add_row(
                "  Docs",
                f"[link=http://{host}:{port}/docs]http://{host}:{port}/docs[/link]",
            )
            table.add_row(
                "  Metrics",
                f"[link=http://{host}:{port}/metrics]http://{host}:{port}/metrics[/link]",
            )

            table.add_section()

            # Quick test commands
            table.add_row("[dim]Quick Test:[/dim]", "")
            table.add_row("", f"[dim]curl http://{host}:{port}/health[/dim]")

            # Find a feature to demo
            if num_features > 0:
                demo_feature = next(iter(store.registry.features.keys()))
                table.add_row(
                    "",
                    f'[dim]curl -X POST http://{host}:{port}/features -d \'{{"entity_id": "u1", "features": ["{demo_feature}"]}}\' -H \'Content-Type: application/json\'[/dim]',
                )

            return Panel(
                table, title="[bold blue]System Status[/bold blue]", border_style="blue"
            )

        layout["main"].update(generate_metrics_table())

        layout["footer"].update(
            Panel(
                f"Dashboard available at: [bold underline]http://{host}:{port}/dashboard[/bold underline] | Press [bold red]Ctrl+C[/bold red] to stop",
                style="dim",
            )
        )

        console.print(f"[green]Successfully loaded features from {file}[/green]")
        console.print(
            "Starting server... (TUI disabled for simple log output compatibility)"
        )

        # NOTE: Running uvicorn inside a Rich Live context is tricky because uvicorn takes over stdout.
        # For this MVP, we will just print the rich header and then run uvicorn.
        # A full TUI requires running uvicorn in a separate thread/process and capturing logs.

        console.print(layout)
        uvicorn.run(app, host=host, port=port)

    except Exception as e:
        console.print(f"[bold red]Error loading features:[/bold red] {e}")
        raise typer.Exit(code=1)


@app.command(name="ui")
def ui(
    file: str = typer.Argument(
        ..., help="Path to the feature definition file (e.g., features.py)"
    ),
    port: int = typer.Option(8501, help="Port to run the UI on"),
    api_port: int = typer.Option(8502, help="Port for API backend"),
) -> None:
    """
    Launches the Fabra UI.

    Starts a Next.js-based UI with a FastAPI backend for exploring
    your Feature Store and Context definitions.

    Example:
        fabra ui features.py
        fabra ui features.py --port 3000
    """
    if not os.path.exists(file):
        console.print(f"[bold red]Error:[/bold red] File '{file}' not found.")
        raise typer.Exit(code=1)

    # Validate FeatureStore exists before starting servers
    sys.path.append(os.getcwd())
    try:
        module_name = os.path.splitext(os.path.basename(file))[0]
        spec = importlib.util.spec_from_file_location(module_name, file)
        if spec is None or spec.loader is None:
            raise ImportError(f"Could not load spec for {file}")

        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)

        store = None
        for attr_name in dir(module):
            attr = getattr(module, attr_name)
            if isinstance(attr, FeatureStore):
                store = attr
                break

        if not store:
            console.print(
                "[bold red]Error:[/bold red] No FeatureStore instance found "
                "in the provided file."
            )
            raise typer.Exit(code=1)
    except Exception as e:
        if "No FeatureStore" in str(e):
            raise
        console.print(f"[bold red]Error:[/bold red] Failed to load module: {e}")
        raise typer.Exit(code=1)

    # Next.js UI with FastAPI backend
    import subprocess  # nosec B404
    import threading
    import time
    import webbrowser

    from .ui_server import run_server

    console.print(
        Panel(
            f"Starting Fabra UI\n\n"
            f"  API Backend: http://127.0.0.1:{api_port}\n"
            f"  UI Frontend: http://localhost:{port}\n\n"
            f"Loading: [bold cyan]{file}[/bold cyan]",
            title="Fabra UI",
            style="bold blue",
        )
    )

    # Check if Next.js build exists
    ui_next_dir = os.path.join(os.path.dirname(__file__), "ui-next")
    if not os.path.exists(os.path.join(ui_next_dir, "node_modules")):
        console.print(
            "[yellow]Warning:[/yellow] Next.js dependencies not installed.\n"
            "Run the following to set up:\n"
            f"  cd {ui_next_dir} && npm install\n\n"
            "Falling back to API-only mode..."
        )
        # Just run the API server
        run_server(file, port=api_port, host="127.0.0.1")
        return

    # Start API server in background thread
    def start_api() -> None:
        import uvicorn
        from .ui_server import load_module, app as api_app

        load_module(file)
        uvicorn.run(api_app, host="127.0.0.1", port=api_port, log_level="warning")

    api_thread = threading.Thread(target=start_api, daemon=True)
    api_thread.start()

    # Wait for API to be ready
    time.sleep(1)

    # Start Next.js dev server
    console.print("[green]Starting UI server...[/green]")
    try:
        # Open browser after a short delay
        def open_browser() -> None:
            time.sleep(3)
            webbrowser.open(f"http://localhost:{port}")

        browser_thread = threading.Thread(target=open_browser, daemon=True)
        browser_thread.start()

        # Run Next.js dev server
        subprocess.run(  # nosec B603 B607
            ["npm", "run", "dev", "--", "-p", str(port)],
            cwd=ui_next_dir,
            check=True,
        )
    except FileNotFoundError:
        console.print(
            "[bold red]Error:[/bold red] npm not found. "
            "Please install Node.js to use the UI."
        )
        raise typer.Exit(code=1)
    except KeyboardInterrupt:
        console.print("\nUI stopped.")


@app.command(name="doctor")
def doctor() -> None:
    """
    Diagnose configuration and connectivity issues.
    """
    from .doctor import run_doctor

    run_doctor()


context_app = typer.Typer(help="Manage and inspect Context Store assemblies.")
app.add_typer(context_app, name="context")


@context_app.command(name="show")
def context_show_cmd(
    context_id: str = typer.Argument(..., help="The Context ID to retrieve"),
    host: str = typer.Option("127.0.0.1", help="Fabra server host"),
    port: int = typer.Option(8000, help="Fabra server port"),
    lineage: bool = typer.Option(
        False, "--lineage", "-l", help="Show only lineage info"
    ),
) -> None:
    """
    Retrieve and display a historical context by ID.

    Example:
      fabra context show 01912345-6789-7abc-def0-123456789abc
      fabra context show <context_id> --lineage
    """
    import urllib.request
    import urllib.error
    import json

    endpoint = "lineage" if lineage else ""
    url = f"http://{host}:{port}/v1/context/{context_id}" + (
        f"/{endpoint}" if endpoint else ""
    )
    console.print(f"Fetching context [bold cyan]{context_id}[/bold cyan] from {url}...")

    if not url.lower().startswith(("http://", "https://")):
        console.print("[bold red]Error:[/bold red] Invalid URL scheme")
        raise typer.Exit(1)

    try:
        req = urllib.request.Request(url)
        api_key = os.getenv("FABRA_API_KEY")
        if api_key:
            req.add_header("X-API-Key", api_key)

        with urllib.request.urlopen(req) as response:  # nosec B310
            if response.status != 200:
                console.print(
                    f"[bold red]Error:[/bold red] Server returned {response.status}"
                )
                raise typer.Exit(1)

            data = json.loads(response.read().decode())

            # Pretty print with Rich
            if lineage:
                title = f"Lineage: {context_id}"
            else:
                title = f"Context: {context_id}"

            console.print(
                Panel(
                    json.dumps(data, indent=2, default=str),
                    title=title,
                    border_style="green",
                )
            )

    except urllib.error.HTTPError as e:
        if e.code == 404:
            console.print(
                f"[bold red]Not Found:[/bold red] Context '{context_id}' does not exist."
            )
        else:
            console.print(f"[bold red]Error:[/bold red] HTTP {e.code}: {e.reason}")
        raise typer.Exit(1)
    except urllib.error.URLError as e:
        console.print(
            f"[bold red]Connection Failed:[/bold red] {e}. Run [bold]fabra doctor[/bold] to check connectivity."
        )
        raise typer.Exit(1)
    except Exception as e:
        console.print(f"[bold red]Error:[/bold red] {e}")
        raise typer.Exit(1)


@context_app.command(name="list")
def context_list_cmd(
    host: str = typer.Option("127.0.0.1", help="Fabra server host"),
    port: int = typer.Option(8000, help="Fabra server port"),
    limit: int = typer.Option(10, "--limit", "-n", help="Number of contexts to list"),
    start: str = typer.Option(None, "--start", "-s", help="Start time (ISO format)"),
    end: str = typer.Option(None, "--end", "-e", help="End time (ISO format)"),
) -> None:
    """
    List recent contexts for debugging/audit.

    Example:
      fabra context list --limit 10
      fabra context list --start 2024-01-01T00:00:00Z --end 2024-01-02T00:00:00Z
    """
    import urllib.request
    import urllib.error
    import json
    from urllib.parse import urlencode
    from typing import Dict, Union

    params: Dict[str, Union[int, str]] = {"limit": limit}
    if start:
        params["start"] = start
    if end:
        params["end"] = end

    url = f"http://{host}:{port}/v1/contexts?{urlencode(params)}"
    console.print(f"Fetching contexts from {url}...")

    if not url.lower().startswith(("http://", "https://")):
        console.print("[bold red]Error:[/bold red] Invalid URL scheme")
        raise typer.Exit(1)

    try:
        req = urllib.request.Request(url)
        api_key = os.getenv("FABRA_API_KEY")
        if api_key:
            req.add_header("X-API-Key", api_key)

        with urllib.request.urlopen(req) as response:  # nosec B310
            if response.status != 200:
                console.print(
                    f"[bold red]Error:[/bold red] Server returned {response.status}"
                )
                raise typer.Exit(1)

            data = json.loads(response.read().decode())

            # Display as a table
            table = Table(title=f"Contexts (limit={limit})", expand=True)
            table.add_column("Context ID", style="cyan", no_wrap=True)
            table.add_column("Timestamp", style="green")
            table.add_column("Version", style="dim")
            table.add_column("Content Preview", style="white", max_width=40)

            contexts = data.get("contexts", [])
            for ctx in contexts:
                content_preview = (
                    ctx.get("content", "")[:40] + "..."
                    if len(ctx.get("content", "")) > 40
                    else ctx.get("content", "")
                )
                table.add_row(
                    ctx.get("context_id", ""),
                    ctx.get("timestamp", ""),
                    ctx.get("version", ""),
                    content_preview,
                )

            console.print(table)
            console.print(f"\n[dim]Total: {len(contexts)} contexts[/dim]")

    except urllib.error.URLError as e:
        console.print(
            f"[bold red]Connection Failed:[/bold red] {e}. Run [bold]fabra doctor[/bold] to check connectivity."
        )
        raise typer.Exit(1)
    except Exception as e:
        console.print(f"[bold red]Error:[/bold red] {e}")
        raise typer.Exit(1)


@context_app.command(name="export")
def context_export_cmd(
    context_id: str = typer.Argument(..., help="The Context ID to export"),
    host: str = typer.Option("127.0.0.1", help="Fabra server host"),
    port: int = typer.Option(8000, help="Fabra server port"),
    format: str = typer.Option(
        "json", "--format", "-f", help="Export format: json, yaml"
    ),
    output: str = typer.Option(
        None, "--output", "-o", help="Output file (default: stdout)"
    ),
) -> None:
    """
    Export a context for audit/debugging.

    Example:
      fabra context export <context_id> --format json
      fabra context export <context_id> --output context.json
      fabra context export <context_id> --format yaml -o context.yaml
    """
    import urllib.request
    import urllib.error
    import json

    url = f"http://{host}:{port}/v1/context/{context_id}"
    console.print(f"Exporting context [bold cyan]{context_id}[/bold cyan]...")

    if not url.lower().startswith(("http://", "https://")):
        console.print("[bold red]Error:[/bold red] Invalid URL scheme")
        raise typer.Exit(1)

    try:
        req = urllib.request.Request(url)
        api_key = os.getenv("FABRA_API_KEY")
        if api_key:
            req.add_header("X-API-Key", api_key)

        with urllib.request.urlopen(req) as response:  # nosec B310
            if response.status != 200:
                console.print(
                    f"[bold red]Error:[/bold red] Server returned {response.status}"
                )
                raise typer.Exit(1)

            data = json.loads(response.read().decode())

            # Format output
            if format == "yaml":
                try:
                    import yaml  # type: ignore[import-untyped]

                    formatted = yaml.dump(
                        data, default_flow_style=False, allow_unicode=True
                    )
                except ImportError:
                    console.print(
                        "[yellow]Warning:[/yellow] PyYAML not installed. Falling back to JSON."
                    )
                    formatted = json.dumps(data, indent=2, default=str)
            else:
                formatted = json.dumps(data, indent=2, default=str)

            # Output to file or stdout
            if output:
                with open(output, "w") as f:
                    f.write(formatted)
                console.print(f"[green]Exported to {output}[/green]")
            else:
                console.print(formatted)

    except urllib.error.HTTPError as e:
        if e.code == 404:
            console.print(
                f"[bold red]Not Found:[/bold red] Context '{context_id}' does not exist."
            )
        else:
            console.print(f"[bold red]Error:[/bold red] HTTP {e.code}: {e.reason}")
        raise typer.Exit(1)
    except urllib.error.URLError as e:
        console.print(
            f"[bold red]Connection Failed:[/bold red] {e}. Run [bold]fabra doctor[/bold] to check connectivity."
        )
        raise typer.Exit(1)
    except Exception as e:
        console.print(f"[bold red]Error:[/bold red] {e}")
        raise typer.Exit(1)


@context_app.command(name="explain")
def explain_cmd(
    ctx_id: str = typer.Argument(..., help="The Context ID to trace"),
    host: str = typer.Option("127.0.0.1", help="Fabra server host"),
    port: int = typer.Option(8000, help="Fabra server port"),
) -> None:
    """
    Fetch and display a context trace (RAG explanation).
    """
    import urllib.request
    import urllib.error
    import json

    url = f"http://{host}:{port}/context/{ctx_id}/explain"
    console.print(f"Fetching trace for [bold cyan]{ctx_id}[/bold cyan] from {url}...")

    if not url.lower().startswith(("http://", "https://")):
        console.print("[bold red]Error:[/bold red] Invalid URL scheme")
        raise typer.Exit(1)

    try:
        # Use standard lib to avoid extra dependencies for CLI
        req = urllib.request.Request(url)
        # Add API Key if present in env
        api_key = os.getenv("FABRA_API_KEY")
        if api_key:
            req.add_header("X-API-Key", api_key)

        with urllib.request.urlopen(req) as response:  # nosec B310
            if response.status != 200:
                console.print(
                    f"[bold red]Error:[/bold red] Server returned {response.status}"
                )
                raise typer.Exit(1)

            data = json.loads(response.read().decode())

            # Pretty print with Rich
            console.print(
                Panel(
                    json.dumps(data, indent=2),
                    title=f"Context Trace: {ctx_id}",
                    border_style="green",
                )
            )

    except urllib.error.URLError as e:
        console.print(
            f"[bold red]Connection Failed:[/bold red] {e}. Run [bold]fabra doctor[/bold] to check connectivity."
        )
        raise typer.Exit(1)
    except Exception as e:
        console.print(f"[bold red]Error:[/bold red] {e}")
        raise typer.Exit(1)


@app.command(name="index")
def index_cmd(
    action: str = typer.Argument(..., help="Action: create, status"),
    name: str = typer.Argument(..., help="Name of the index"),
    dimension: int = typer.Option(1536, help="Vector dimension (create only)"),
    postgres_url: str = typer.Option(
        None, envvar="FABRA_POSTGRES_URL", help="Postgres URL Override"
    ),
    dry_run: bool = typer.Option(
        False, "--dry-run", "-n", help="Preview changes without executing"
    ),
) -> None:
    """
    Manage vector indexes.
    Usage:
      fabra index create my_index --dimension=1536
      fabra index status my_index
      fabra index create my_index --dry-run  # Preview only
    """
    import asyncio
    from .store.postgres import PostgresOfflineStore
    from sqlalchemy import text

    url = postgres_url or os.getenv("FABRA_POSTGRES_URL")
    if not url:
        console.print("[bold red]Error:[/bold red] FABRA_POSTGRES_URL not set.")
        raise typer.Exit(1)

    async def run_action() -> None:
        try:
            store = PostgresOfflineStore(url)

            if action == "create":
                table_name = f"fabra_index_{name}"
                if dry_run:
                    console.print(
                        Panel(
                            f"[bold]Dry Run:[/bold] Would create index table\n\n"
                            f"  Table:     [cyan]{table_name}[/cyan]\n"
                            f"  Dimension: [cyan]{dimension}[/cyan]\n"
                            f"  Database:  [dim]{url.split('@')[-1] if '@' in url else url}[/dim]\n\n"
                            f"Run without --dry-run to execute.",
                            title="Index Preview",
                            style="yellow",
                        )
                    )
                    return

                console.print(
                    f"Creating index [bold]{name}[/bold] (dim={dimension})..."
                )
                await store.create_index_table(name, dimension)
                console.print(f"[green]Index '{name}' created successfully.[/green]")

            elif action == "status":
                table = f"fabra_index_{name}"
                async with store.engine.connect() as conn:  # type: ignore[no-untyped-call]
                    # Check if exists
                    exists = await conn.execute(
                        text(f"SELECT to_regclass('public.{table}')")
                    )
                    if not exists.scalar():
                        console.print(
                            f"[red]Index table '{table}' does not exist.[/red]"
                        )
                        return

                    # Count rows
                    res = await conn.execute(text(f"SELECT COUNT(*) FROM {table}"))  # nosec
                    count = res.scalar()
                    console.print(f"Index: [bold]{name}[/bold]")
                    console.print(f"Table: {table}")
                    console.print(f"Rows:  [cyan]{count}[/cyan]")

            else:
                console.print(f"[red]Unknown action: {action}[/red]")

            await store.engine.dispose()  # type: ignore[no-untyped-call]

        except Exception as e:
            console.print(f"[bold red]Error:[/bold red] {e}")
            raise typer.Exit(1)

    asyncio.run(run_action())


@app.command(name="deploy")
def deploy_cmd(
    target: str = typer.Argument(
        "fly", help="Deployment target: fly, cloudrun, ecs, render, railway"
    ),
    file: str = typer.Option(
        "features.py", "--file", "-f", help="Feature definition file"
    ),
    app_name: str = typer.Option("fabra-app", "--name", "-n", help="Application name"),
    region: str = typer.Option("iad", "--region", "-r", help="Deployment region"),
    output: str = typer.Option(".", "--output", "-o", help="Output directory"),
    dry_run: bool = typer.Option(
        False, "--dry-run", help="Preview generated files without writing"
    ),
) -> None:
    """
    Generate deployment configuration for cloud platforms.

    Usage:
      fabra deploy fly --name my-app          # Generate fly.toml
      fabra deploy cloudrun --name my-app     # Generate Cloud Run config
      fabra deploy ecs --name my-app          # Generate ECS task definition
      fabra deploy --dry-run                  # Preview without writing
    """
    # Common Dockerfile for all deployments
    dockerfile = f"""# Auto-generated by Fabra CLI
FROM python:3.11-slim

WORKDIR /app

# Install dependencies
COPY requirements.txt .
RUN pip install --no-cache-dir -r requirements.txt

# Copy application
COPY {file} .
COPY . .

# Expose port
EXPOSE 8000

# Health check
HEALTHCHECK --interval=30s --timeout=3s \\
  CMD curl -f http://localhost:8000/health || exit 1

# Run server
CMD ["fabra", "serve", "{file}", "--host", "0.0.0.0", "--port", "8000"]
"""

    requirements_txt = """fabra>=2.2.0
redis>=4.0.0
asyncpg>=0.27.0
"""

    configs: dict[str, dict[str, str]] = {
        "fly": {
            "fly.toml": f"""# Auto-generated by Fabra CLI
app = "{app_name}"
primary_region = "{region}"

[build]
  dockerfile = "Dockerfile"

[env]
  PORT = "8000"
  FABRA_ENV = "production"

[http_service]
  internal_port = 8000
  force_https = true
  auto_stop_machines = true
  auto_start_machines = true
  min_machines_running = 1

[[services]]
  internal_port = 8000
  protocol = "tcp"

  [[services.ports]]
    handlers = ["http"]
    port = 80

  [[services.ports]]
    handlers = ["tls", "http"]
    port = 443

  [[services.http_checks]]
    interval = "10s"
    timeout = "2s"
    path = "/health"
""",
        },
        "cloudrun": {
            "service.yaml": f"""# Auto-generated by Fabra CLI
apiVersion: serving.knative.dev/v1
kind: Service
metadata:
  name: {app_name}
spec:
  template:
    metadata:
      annotations:
        autoscaling.knative.dev/minScale: "1"
        autoscaling.knative.dev/maxScale: "10"
    spec:
      containers:
        - image: gcr.io/PROJECT_ID/{app_name}:latest
          ports:
            - containerPort: 8000
          env:
            - name: FABRA_ENV
              value: production
          resources:
            limits:
              cpu: "1"
              memory: 512Mi
          startupProbe:
            httpGet:
              path: /health
              port: 8000
            initialDelaySeconds: 5
            periodSeconds: 10
""",
            "cloudbuild.yaml": f"""# Auto-generated by Fabra CLI
steps:
  - name: gcr.io/cloud-builders/docker
    args: ["build", "-t", "gcr.io/$PROJECT_ID/{app_name}:$COMMIT_SHA", "."]
  - name: gcr.io/cloud-builders/docker
    args: ["push", "gcr.io/$PROJECT_ID/{app_name}:$COMMIT_SHA"]
  - name: gcr.io/google.com/cloudsdktool/cloud-sdk
    entrypoint: gcloud
    args:
      - run
      - deploy
      - {app_name}
      - --image=gcr.io/$PROJECT_ID/{app_name}:$COMMIT_SHA
      - --region={region}
      - --platform=managed
images:
  - gcr.io/$PROJECT_ID/{app_name}:$COMMIT_SHA
""",
        },
        "ecs": {
            "task-definition.json": f"""{{
  "family": "{app_name}",
  "networkMode": "awsvpc",
  "requiresCompatibilities": ["FARGATE"],
  "cpu": "256",
  "memory": "512",
  "executionRoleArn": "arn:aws:iam::ACCOUNT_ID:role/ecsTaskExecutionRole",
  "containerDefinitions": [
    {{
      "name": "{app_name}",
      "image": "ACCOUNT_ID.dkr.ecr.{region}.amazonaws.com/{app_name}:latest",
      "essential": true,
      "portMappings": [
        {{
          "containerPort": 8000,
          "protocol": "tcp"
        }}
      ],
      "environment": [
        {{"name": "FABRA_ENV", "value": "production"}}
      ],
      "healthCheck": {{
        "command": ["CMD-SHELL", "curl -f http://localhost:8000/health || exit 1"],
        "interval": 30,
        "timeout": 5,
        "retries": 3
      }},
      "logConfiguration": {{
        "logDriver": "awslogs",
        "options": {{
          "awslogs-group": "/ecs/{app_name}",
          "awslogs-region": "{region}",
          "awslogs-stream-prefix": "ecs"
        }}
      }}
    }}
  ]
}}
""",
        },
        "render": {
            "render.yaml": f"""# Auto-generated by Fabra CLI
services:
  - type: web
    name: {app_name}
    env: docker
    region: {region}
    plan: starter
    healthCheckPath: /health
    envVars:
      - key: FABRA_ENV
        value: production
      - key: PORT
        value: 8000
""",
        },
        "railway": {
            "railway.json": f"""{{
  "$schema": "https://railway.app/railway.schema.json",
  "build": {{
    "builder": "DOCKERFILE",
    "dockerfilePath": "Dockerfile"
  }},
  "deploy": {{
    "startCommand": "fabra serve {file} --host 0.0.0.0 --port $PORT",
    "healthcheckPath": "/health",
    "healthcheckTimeout": 30,
    "restartPolicyType": "ON_FAILURE",
    "restartPolicyMaxRetries": 3
  }}
}}
""",
        },
    }

    if target not in configs:
        console.print(
            f"[bold red]Unknown target:[/bold red] {target}\n"
            f"Available targets: {', '.join(configs.keys())}"
        )
        raise typer.Exit(1)

    # Get target-specific files
    target_files = configs[target]

    # Add common files
    all_files = {
        "Dockerfile": dockerfile,
        "requirements.txt": requirements_txt,
        **target_files,
    }

    if dry_run:
        console.print(
            Panel(
                f"[bold]Dry Run:[/bold] Would create {len(all_files)} files for {target}\n\n"
                + "\n".join(f"  [cyan]{f}[/cyan]" for f in all_files.keys())
                + "\n\nRun without --dry-run to create files.",
                title=f"Deploy to {target.title()}",
                style="yellow",
            )
        )
        # Show a preview of each file
        for filename, content in all_files.items():
            console.print(f"\n[bold]{filename}:[/bold]")
            console.print(
                Panel(content.strip()[:500] + ("..." if len(content) > 500 else ""))
            )
        return

    # Create output directory if needed
    if not os.path.exists(output):
        os.makedirs(output)
        console.print(f"Created directory: [bold cyan]{output}[/bold cyan]")

    # Write files
    created_files = []
    for filename, content in all_files.items():
        filepath = os.path.join(output, filename)
        if os.path.exists(filepath):
            console.print(f"[yellow]Warning:[/yellow] {filepath} exists. Skipping.")
        else:
            with open(filepath, "w") as f:
                f.write(content.strip())
            created_files.append(filename)
            console.print(f"Created [bold]{filename}[/bold]")

    # Show next steps
    console.print("\n[green]Deployment files generated![/green]\n")

    next_steps = {
        "fly": [
            "fly auth login",
            "fly launch --no-deploy",
            "fly secrets set FABRA_REDIS_URL=... FABRA_POSTGRES_URL=...",
            "fly deploy",
        ],
        "cloudrun": [
            "gcloud auth login",
            "gcloud config set project YOUR_PROJECT",
            "gcloud builds submit",
        ],
        "ecs": [
            "aws ecr create-repository --repository-name " + app_name,
            "docker build -t " + app_name + " .",
            "docker push ACCOUNT_ID.dkr.ecr.REGION.amazonaws.com/" + app_name,
            "aws ecs register-task-definition --cli-input-json file://task-definition.json",
        ],
        "render": [
            "Connect your GitHub repo to Render",
            "Render will auto-deploy on push",
        ],
        "railway": [
            "railway login",
            "railway init",
            "railway up",
        ],
    }

    console.print("[bold]Next steps:[/bold]")
    for i, step in enumerate(next_steps.get(target, []), 1):
        console.print(f"  {i}. [dim]{step}[/dim]")


if __name__ == "__main__":
    app()



from abc import ABC, abstractmethod
from typing import Optional, Any
import structlog
import functools

logger = structlog.get_logger()


class TokenCounter(ABC):
    @abstractmethod
    def count(self, text: str) -> int:
        """Count tokens in text."""
        pass


class OpenAITokenCounter(TokenCounter):
    def __init__(self, model: str = "gpt-4o"):
        self.model = model
        self.encoder: Optional[Any] = None
        try:
            import tiktoken

            self.encoder = tiktoken.encoding_for_model(model)
        except Exception as e:
            logger.warning(f"Failed to load tiktoken for {model}: {e}")
            self.encoder = None

    def count(self, text: str) -> int:
        if not self.encoder:
            # Fallback estimation: chars / 4
            return len(text) // 4
        try:
            return len(self.encoder.encode(text))
        except Exception as e:
            logger.error(f"Token count error: {e}")
            return len(text) // 4


class AnthropicTokenCounter(TokenCounter):
    def __init__(self, model: str = "claude-3-5-sonnet-20240620"):
        self.model = model
        self.client = None
        try:
            from anthropic import Anthropic

            self.client = Anthropic()
        except Exception as e:
            logger.warning(f"Failed to load anthropic SDK: {e}")

    def count(self, text: str) -> int:
        if not self.client:
            return len(text) // 4

        return _get_anthropic_token_count(self.model, text)


@functools.lru_cache(maxsize=1024)
def _get_anthropic_token_count(model: str, text: str) -> int:
    try:
        # We instantiate client here (stateless mostly) to avoid passing 'self'
        from anthropic import Anthropic

        client = Anthropic()
        response = client.beta.messages.count_tokens(
            model=model, messages=[{"role": "user", "content": text}]
        )
        return response.input_tokens
    except Exception as e:
        logger.error(f"Anthropic token count error: {e}")
        return len(text) // 4



"""Tests for Freshness SLA feature (v1.5)."""

import pytest
from datetime import datetime, timezone, timedelta

from fabra.utils.time import (
    parse_duration_to_ms,
    format_ms_to_human,
    validate_sla,
    InvalidSLAFormatError,
)
from fabra.exceptions import FreshnessSLAError
from fabra.context import context, ContextItem, AssemblyTracker


class TestSLAParsing:
    """Tests for SLA duration parsing (2.1)."""

    def test_parse_seconds(self):
        assert parse_duration_to_ms("30s") == 30_000
        assert parse_duration_to_ms("1s") == 1_000
        assert parse_duration_to_ms("120s") == 120_000

    def test_parse_minutes(self):
        assert parse_duration_to_ms("5m") == 300_000
        assert parse_duration_to_ms("1m") == 60_000
        assert parse_duration_to_ms("30m") == 1_800_000

    def test_parse_hours(self):
        assert parse_duration_to_ms("1h") == 3_600_000
        assert parse_duration_to_ms("24h") == 86_400_000

    def test_parse_days(self):
        assert parse_duration_to_ms("1d") == 86_400_000
        assert parse_duration_to_ms("7d") == 604_800_000

    def test_parse_milliseconds(self):
        assert parse_duration_to_ms("500ms") == 500
        assert parse_duration_to_ms("1000ms") == 1_000

    def test_parse_with_decimals(self):
        assert parse_duration_to_ms("1.5h") == 5_400_000
        assert parse_duration_to_ms("2.5m") == 150_000

    def test_parse_case_insensitive(self):
        assert parse_duration_to_ms("5M") == 300_000
        assert parse_duration_to_ms("1H") == 3_600_000

    def test_parse_with_whitespace(self):
        assert parse_duration_to_ms(" 5m ") == 300_000

    def test_invalid_format_raises(self):
        with pytest.raises(InvalidSLAFormatError):
            parse_duration_to_ms("invalid")

        with pytest.raises(InvalidSLAFormatError):
            parse_duration_to_ms("5x")

        with pytest.raises(InvalidSLAFormatError):
            parse_duration_to_ms("")

        with pytest.raises(InvalidSLAFormatError):
            parse_duration_to_ms("m5")

    def test_validate_sla_none(self):
        assert validate_sla(None) is None

    def test_validate_sla_valid(self):
        assert validate_sla("5m") == 300_000


class TestFormatMsToHuman:
    """Tests for human-readable duration formatting."""

    def test_format_milliseconds(self):
        assert format_ms_to_human(500) == "500ms"
        assert format_ms_to_human(999) == "999ms"

    def test_format_seconds(self):
        assert format_ms_to_human(1000) == "1s"
        assert format_ms_to_human(30000) == "30s"

    def test_format_minutes(self):
        assert format_ms_to_human(60000) == "1m"
        assert format_ms_to_human(90000) == "1m 30s"

    def test_format_hours(self):
        assert format_ms_to_human(3600000) == "1h"
        assert format_ms_to_human(5400000) == "1h 30m"

    def test_format_days(self):
        assert format_ms_to_human(86400000) == "1d"
        assert format_ms_to_human(90000000) == "1d 1h"


class TestFreshnessSLAError:
    """Tests for FreshnessSLAError exception."""

    def test_error_creation(self):
        violations = [
            {"feature": "user_tier", "age_ms": 360000, "sla_ms": 300000},
            {"feature": "purchase_count", "age_ms": 400000, "sla_ms": 300000},
        ]
        error = FreshnessSLAError("SLA breached", violations)

        assert error.message == "SLA breached"
        assert len(error.violations) == 2
        assert error.violations[0]["feature"] == "user_tier"

    def test_error_str(self):
        violations = [{"feature": "user_tier", "age_ms": 360000, "sla_ms": 300000}]
        error = FreshnessSLAError("SLA breached", violations)

        assert "user_tier" in str(error)
        assert "360000ms" in str(error)
        assert "300000ms" in str(error)

    def test_error_repr(self):
        violations = [{"feature": "user_tier", "age_ms": 360000, "sla_ms": 300000}]
        error = FreshnessSLAError("SLA breached", violations)

        assert "FreshnessSLAError" in repr(error)
        assert "user_tier" in repr(error)


class TestContextDecoratorWithFreshnessSLA:
    """Tests for @context decorator with freshness_sla parameter."""

    @pytest.mark.asyncio
    async def test_invalid_sla_format_raises_at_decoration(self):
        """Invalid SLA format should raise immediately when decorator is applied."""
        with pytest.raises(InvalidSLAFormatError):

            @context(freshness_sla="invalid")
            async def bad_context():
                return [ContextItem(content="test")]

    @pytest.mark.asyncio
    async def test_valid_sla_format_accepted(self):
        """Valid SLA formats should be accepted."""

        @context(freshness_sla="5m")
        async def good_context():
            return [ContextItem(content="test")]

        # Should not raise
        assert good_context is not None

    @pytest.mark.asyncio
    async def test_context_freshness_guaranteed_when_all_fresh(self):
        """Context should be 'guaranteed' when all features are within SLA."""

        @context(freshness_sla="5m")
        async def fresh_context():
            return [ContextItem(content="test")]

        ctx = await fresh_context()
        assert ctx.meta["freshness_status"] == "guaranteed"
        assert ctx.meta["freshness_violations"] == []

    @pytest.mark.asyncio
    async def test_context_without_sla_is_guaranteed(self):
        """Context without SLA should default to 'guaranteed'."""

        @context()
        async def no_sla_context():
            return [ContextItem(content="test")]

        ctx = await no_sla_context()
        assert ctx.meta["freshness_status"] == "guaranteed"


class TestContextDecoratorStrictMode:
    """Tests for strict mode with FreshnessSLAError."""

    @pytest.mark.asyncio
    async def test_strict_mode_default_is_false(self):
        """Default freshness_strict should be False."""

        @context(freshness_sla="5m")
        async def context_func():
            return [ContextItem(content="test")]

        # Should not raise even if violations exist (mock would be needed)
        ctx = await context_func()
        assert ctx is not None

    @pytest.mark.asyncio
    async def test_strict_mode_no_error_when_fresh(self):
        """Strict mode should not raise when all features are fresh."""

        @context(freshness_sla="5m", freshness_strict=True)
        async def strict_context():
            return [ContextItem(content="test")]

        ctx = await strict_context()
        assert ctx.meta["freshness_status"] == "guaranteed"


class TestAssemblyTrackerFreshness:
    """Tests for AssemblyTracker freshness tracking."""

    def test_get_stalest_feature_ms_empty(self):
        """Empty tracker should return 0 for stalest feature."""
        tracker = AssemblyTracker(context_id="test-id")
        assert tracker.get_stalest_feature_ms() == 0

    def test_get_stalest_feature_ms_with_features(self):
        """Tracker should return max freshness_ms."""
        tracker = AssemblyTracker(context_id="test-id")

        # Record features with different ages
        now = datetime.now(timezone.utc)
        tracker.record_feature(
            feature_name="feature1",
            entity_id="user_1",
            value="val1",
            timestamp=now - timedelta(seconds=30),
            source="compute",
        )
        tracker.record_feature(
            feature_name="feature2",
            entity_id="user_1",
            value="val2",
            timestamp=now - timedelta(seconds=60),
            source="compute",
        )

        stalest = tracker.get_stalest_feature_ms()
        # Should be approximately 60 seconds = 60000ms
        assert stalest > 55000
        assert stalest < 65000

    def test_record_feature_calculates_freshness(self):
        """record_feature should calculate freshness_ms from timestamp."""
        tracker = AssemblyTracker(context_id="test-id")

        # Record a feature that's 5 seconds old
        old_time = datetime.now(timezone.utc) - timedelta(seconds=5)
        tracker.record_feature(
            feature_name="test_feature",
            entity_id="entity_1",
            value="value",
            timestamp=old_time,
            source="cache",
        )

        assert len(tracker.features) == 1
        assert tracker.features[0].feature_name == "test_feature"
        assert tracker.features[0].freshness_ms > 4000
        assert tracker.features[0].freshness_ms < 6000


class TestContextMetaFreshnessFields:
    """Tests for context meta freshness fields."""

    @pytest.mark.asyncio
    async def test_context_includes_freshness_violations_field(self):
        """Context meta should include freshness_violations list."""

        @context(freshness_sla="5m")
        async def context_func():
            return [ContextItem(content="test")]

        ctx = await context_func()
        assert "freshness_violations" in ctx.meta
        assert isinstance(ctx.meta["freshness_violations"], list)

    @pytest.mark.asyncio
    async def test_context_includes_freshness_sla_ms_field(self):
        """Context meta should include freshness_sla_ms when SLA is set."""

        @context(freshness_sla="5m")
        async def context_func():
            return [ContextItem(content="test")]

        ctx = await context_func()
        assert ctx.meta["freshness_sla_ms"] == 300_000

    @pytest.mark.asyncio
    async def test_context_freshness_sla_ms_none_when_not_set(self):
        """Context meta should have freshness_sla_ms=None when no SLA."""

        @context()
        async def context_func():
            return [ContextItem(content="test")]

        ctx = await context_func()
        assert ctx.meta["freshness_sla_ms"] is None



"""
Recommendation Engine Features

This example demonstrates how to build a product recommendation system
using Fabra's feature store. Features include:
- User behavioral features (browsing patterns, purchase history)
- Product features (popularity, category embeddings)
- Cross-entity features (user-product affinity scores)
"""

from datetime import timedelta
from fabra.core import FeatureStore, entity, feature
from fabra.store import DuckDBOfflineStore, InMemoryOnlineStore

# Initialize Store
store = FeatureStore(
    offline_store=DuckDBOfflineStore(), online_store=InMemoryOnlineStore()
)


# --- Entities ---


@entity(store)
class User:
    user_id: str


@entity(store)
class Product:
    product_id: str


# --- User Features ---


@feature(entity=User, refresh=timedelta(minutes=5), materialize=True)
def user_browsing_category(user_id: str) -> str:
    """Most viewed product category in the last hour."""
    # In production: aggregate from clickstream data
    categories = ["electronics", "clothing", "home", "sports"]
    return categories[hash(user_id) % len(categories)]


@feature(entity=User, refresh=timedelta(hours=1), materialize=True)
def user_purchase_count_30d(user_id: str) -> int:
    """Number of purchases in the last 30 days."""
    # In production: COUNT(*) from orders WHERE created_at > now() - interval '30 days'
    return (hash(user_id) % 20) + 1


@feature(entity=User, refresh=timedelta(hours=24), materialize=True)
def user_avg_order_value(user_id: str) -> float:
    """Average order value across all purchases."""
    # In production: AVG(total) from orders
    base = 50 + (hash(user_id) % 200)
    return round(base + (hash(user_id + "aov") % 100) / 10, 2)


@feature(entity=User, refresh=timedelta(minutes=30), materialize=True)
def user_price_sensitivity(user_id: str) -> str:
    """Price sensitivity segment: low, medium, high."""
    # In production: ML model output based on coupon usage, cart abandonment, etc.
    segments = ["low", "medium", "high"]
    return segments[hash(user_id + "price") % len(segments)]


# --- Product Features ---


@feature(entity=Product, refresh=timedelta(hours=1), materialize=True)
def product_popularity_score(product_id: str) -> float:
    """Popularity score (0-100) based on views, purchases, and ratings."""
    # In production: composite score from multiple signals
    return round((hash(product_id) % 100) + (hash(product_id + "pop") % 10) / 10, 2)


@feature(entity=Product, refresh=timedelta(hours=6), materialize=True)
def product_inventory_level(product_id: str) -> str:
    """Inventory status: in_stock, low_stock, out_of_stock."""
    # In production: real-time from inventory system
    levels = ["in_stock", "in_stock", "in_stock", "low_stock", "out_of_stock"]
    return levels[hash(product_id + "inv") % len(levels)]


@feature(entity=Product, refresh=timedelta(hours=24), materialize=True)
def product_rating(product_id: str) -> float:
    """Average rating (1-5 stars)."""
    # In production: AVG(rating) from reviews
    base = 3.0 + (hash(product_id) % 20) / 10
    return round(min(5.0, base), 1)


# --- Pre-load demo data ---
if __name__ == "__main__":
    import asyncio

    async def load_demo_data() -> None:
        # Load sample user features
        await store.online_store.set_online_features(
            "User",
            "user_001",
            {
                "user_browsing_category": "electronics",
                "user_purchase_count_30d": 12,
                "user_avg_order_value": 175.50,
                "user_price_sensitivity": "medium",
            },
        )

        await store.online_store.set_online_features(
            "User",
            "user_002",
            {
                "user_browsing_category": "clothing",
                "user_purchase_count_30d": 3,
                "user_avg_order_value": 65.00,
                "user_price_sensitivity": "high",
            },
        )

        # Load sample product features
        products = ["prod_100", "prod_101", "prod_102"]
        for pid in products:
            await store.online_store.set_online_features(
                "Product",
                pid,
                {
                    "product_popularity_score": round(70 + hash(pid) % 30, 2),
                    "product_inventory_level": "in_stock",
                    "product_rating": round(4.0 + (hash(pid) % 10) / 10, 1),
                },
            )

        print("Demo data loaded successfully!")

    asyncio.run(load_demo_data())



import pytest
from typing import List, Dict, Any
from unittest.mock import MagicMock
from fabra.core import FeatureStore
from fabra.hooks import Hook
from fabra.store import InMemoryOnlineStore


class SpyHook(Hook):
    def __init__(self) -> None:
        self.before_args: Any = None
        self.after_args: Any = None

    async def before_feature_retrieval(
        self, entity_name: str, entity_id: str, features: List[str]
    ) -> None:
        self.before_args = (entity_name, entity_id, features)

    async def after_feature_retrieval(
        self,
        entity_name: str,
        entity_id: str,
        features: List[str],
        result: Dict[str, Any],
    ) -> None:
        self.after_args = (entity_name, entity_id, features, result)
        # Modify result
        result["spy_value"] = 999


@pytest.mark.asyncio
async def test_feature_store_hooks_integration() -> None:
    spy = SpyHook()
    # Mock offline store to avoid errors
    store = FeatureStore(
        offline_store=MagicMock(), online_store=InMemoryOnlineStore(), hooks=[spy]
    )

    # Register a dummy feature
    def dummy_func(id: str) -> int:
        return 42

    # Register entity imperatively
    store.register_entity("User", "user_id")

    store.register_feature("age", "User", dummy_func)

    # Call get_online_features
    # This will use compute path because cache is empty
    result = await store.get_online_features("User", "u1", ["age"])

    # Check Before Hook
    assert spy.before_args == ("User", "u1", ["age"])

    # Check After Hook
    assert spy.after_args is not None
    assert spy.after_args[3]["age"] == 42

    # Check if hook modification persisted in return value
    # The hook modifies the result dict in place
    assert result["spy_value"] == 999



from prometheus_client import Counter, Histogram
import time
from typing import Optional, Any

# --- Metric Definitions ---

CONTEXT_ASSEMBLY_TOTAL = Counter(
    "fabra_context_assembly_total",
    "Total number of context assembly operations",
    ["name", "status"],
)

CONTEXT_LATENCY_SECONDS = Histogram(
    "fabra_context_latency_seconds",
    "Latency of context assembly operations",
    ["name"],
)

CONTEXT_TOKENS_TOTAL = Counter(
    "fabra_context_tokens_total", "Total tokens generated in contexts", ["name"]
)

CONTEXT_CACHE_HIT_TOTAL = Counter(
    "fabra_context_cache_hit_total",
    "Total number of cache hits for context assembly",
    ["name"],
)

INDEX_WRITE_TOTAL = Counter(
    "fabra_index_write_total",
    "Total number of documents written to vector index",
    ["index_name"],
)

# --- Freshness SLA Metrics (v1.5) ---

CONTEXT_FRESHNESS_STATUS_TOTAL = Counter(
    "fabra_context_freshness_status_total",
    "Total contexts by freshness status",
    ["name", "status"],  # status: "guaranteed" or "degraded"
)

CONTEXT_FRESHNESS_VIOLATIONS_TOTAL = Counter(
    "fabra_context_freshness_violations_total",
    "Total freshness SLA violations by feature",
    ["name", "feature"],
)

CONTEXT_STALEST_FEATURE_SECONDS = Histogram(
    "fabra_context_stalest_feature_seconds",
    "Age of the stalest feature used in context assembly",
    ["name"],
    buckets=(0.1, 0.5, 1, 5, 10, 30, 60, 300, 600, 1800, 3600),
)


class ContextMetrics:
    """Helper to track context metrics."""

    def __init__(self, name: str) -> None:
        self.name = name
        self.start_time = 0.0

    def __enter__(self) -> "ContextMetrics":
        self.start_time = time.time()
        return self

    def __exit__(
        self,
        exc_type: Optional[type],
        exc_val: Optional[BaseException],
        exc_tb: Optional[Any],
    ) -> None:
        duration = time.time() - self.start_time
        status = "failure" if exc_type else "success"

        CONTEXT_ASSEMBLY_TOTAL.labels(name=self.name, status=status).inc()
        CONTEXT_LATENCY_SECONDS.labels(name=self.name).observe(duration)

    def record_tokens(self, count: int) -> None:
        CONTEXT_TOKENS_TOTAL.labels(name=self.name).inc(count)

    def record_cache_hit(self) -> None:
        CONTEXT_CACHE_HIT_TOTAL.labels(name=self.name).inc()

    def record_freshness_status(self, status: str) -> None:
        """Record freshness status (guaranteed or degraded)."""
        CONTEXT_FRESHNESS_STATUS_TOTAL.labels(name=self.name, status=status).inc()

    def record_freshness_violation(self, feature: str) -> None:
        """Record a freshness SLA violation for a specific feature."""
        CONTEXT_FRESHNESS_VIOLATIONS_TOTAL.labels(name=self.name, feature=feature).inc()

    def record_stalest_feature(self, age_seconds: float) -> None:
        """Record the age of the stalest feature in seconds."""
        CONTEXT_STALEST_FEATURE_SECONDS.labels(name=self.name).observe(age_seconds)



import os
import json
import structlog
from typing import Dict, TypedDict, Optional
from datetime import datetime, timedelta

logger = structlog.get_logger()

# Version of the pricing table. Format: YYYY-MM
PRICING_VERSION = "2025-01"


class PriceRates(TypedDict):
    input: float
    output: float


# Pricing Table (USD per 1M tokens)
# Sources checked Dec 2024
DEFAULT_PRICING: Dict[str, PriceRates] = {
    # Anthropic
    "claude-4.5-opus": {"input": 5.00, "output": 25.00},
    "claude-4.5-sonnet": {"input": 3.00, "output": 15.00},
    "claude-4.5-haiku": {"input": 1.00, "output": 5.00},
    # Cohere
    "command-a": {"input": 2.50, "output": 10.00},
    "command-r": {"input": 0.15, "output": 0.60},
    "embed-4": {"input": 0.12, "output": 0.0},
    # OpenAI
    "text-embedding-3-small": {"input": 0.02, "output": 0.0},
    "text-embedding-3-large": {"input": 0.13, "output": 0.0},
    "text-embedding-ada-002": {"input": 0.10, "output": 0.0},
    "gpt-4.1": {"input": 2.00, "output": 8.00},
    "gpt-4.1-mini": {"input": 0.40, "output": 1.60},
    "gpt-5": {"input": 1.25, "output": 10.00},
    "gpt-5-pro": {"input": 15.00, "output": 120.00},
    "gpt-4o": {"input": 2.50, "output": 10.00},
    # Fallback
    "default": {"input": 5.00, "output": 15.00},
}

_merged_pricing: Optional[Dict[str, PriceRates]] = None
_checked_freshness = False


def _get_pricing() -> Dict[str, PriceRates]:
    """
    Returns the pricing table, merging default with optional custom file.
    Only loads once.
    """
    global _merged_pricing, _checked_freshness

    # 1. Freshness Check (Once)
    if not _checked_freshness:
        _check_pricing_freshness()
        _checked_freshness = True

    if _merged_pricing is not None:
        return _merged_pricing

    # Start with defaults
    pricing = DEFAULT_PRICING.copy()

    # 2. Check for Override File
    custom_file = os.environ.get("FABRA_CUSTOM_PRICING_FILE")
    if custom_file:
        try:
            if os.path.exists(custom_file):
                with open(custom_file, "r") as f:
                    custom_data = json.load(f)
                    # Merge logic: overwrite keys
                    for k, v in custom_data.items():
                        if "input" in v and "output" in v:
                            pricing[k] = {
                                "input": float(v["input"]),
                                "output": float(v["output"]),
                            }
                logger.info("loaded_custom_pricing", file=custom_file)
            else:
                logger.warning("custom_pricing_file_not_found", file=custom_file)
        except Exception as e:
            logger.error("failed_to_load_custom_pricing", error=str(e))

    _merged_pricing = pricing
    return pricing


def _check_pricing_freshness() -> None:
    """Logs a warning if the internal pricing version is > 6 months old."""
    try:
        version_date = datetime.strptime(PRICING_VERSION, "%Y-%m")
        # Check if > 180 days (approx 6 months)
        if datetime.now() - version_date > timedelta(days=180):
            logger.warning(
                "pricing_data_stale",
                version=PRICING_VERSION,
                msg="Pricing data may be outdated (>6 months). Set FABRA_CUSTOM_PRICING_FILE or upgrade Fabra.",
            )
    except Exception:  # nosec
        pass


def estimate_cost(model: str, input_tokens: int, output_tokens: int = 0) -> float:
    """
    Estimates the cost of a request based on the model and token counts.

    Args:
        model: Model identifier (e.g., 'gpt-4o'). Defaults to 'default' pricing if unknown.
        input_tokens: Number of input/context tokens.
        output_tokens: Number of output/generated tokens.

    Returns:
        Estimated cost in USD.
    """
    pricing = _get_pricing()

    rates = pricing.get(model)
    if not rates:
        # Try finding a matching prefix or fallback
        # e.g. "gpt-4o-2024..." might match "gpt-4o" if we wanted fuzzy matching,
        # but exact match + default is safer for now.
        rates = pricing["default"]

    cost = (input_tokens / 1_000_000 * rates["input"]) + (
        output_tokens / 1_000_000 * rates["output"]
    )
    return round(cost, 6)



import pytest
from unittest.mock import MagicMock
from fabra.context import context, ContextItem


@pytest.mark.asyncio
async def test_context_budget_graceful() -> None:
    mock_counter = MagicMock()
    mock_counter.count.side_effect = lambda x: len(x)

    # Define context that will overflow budget with required items
    # Budget 5. "Required" is 8 chars.
    @context(name="overflow_ctx", max_tokens=5, token_counter=mock_counter)
    async def overflow_func() -> list[ContextItem]:
        return [
            ContextItem(content="Required", required=True)  # 8 chars > 5
        ]

    # Should NOT raise ContextBudgetError
    ctx = await overflow_func()

    assert ctx.content == "Required"
    assert ctx.meta["budget_exceeded"] is True
    assert ctx.meta["dropped_items"] == 0


@pytest.mark.asyncio
async def test_context_budget_graceful_string() -> None:
    mock_counter = MagicMock()
    mock_counter.count.side_effect = lambda x: len(x)

    @context(name="overflow_str", max_tokens=5, token_counter=mock_counter)
    async def overflow_str_func() -> str:
        return "TooLong"  # 7 chars > 5

    ctx = await overflow_str_func()
    assert ctx.content == "TooLong"
    assert ctx.meta["budget_exceeded"] is True



from __future__ import annotations
from unittest.mock import MagicMock, patch
from fabra.utils.tokens import OpenAITokenCounter, AnthropicTokenCounter


def test_openai_token_counter_basic() -> None:
    # We expect tiktoken to be installed
    counter = OpenAITokenCounter()
    text = "Hello world"
    count = counter.count(text)
    # tiktoken: "Hello" (1) + " world" (1) = 2 usually
    assert count > 0


def test_anthropic_token_counter_mocked() -> None:
    # Patch the global anthropic class since it is imported inside the method
    # but refers to the installed package.
    with patch("anthropic.Anthropic") as MockAnthropic:
        mock_client = MagicMock()
        MockAnthropic.return_value = mock_client

        mock_response = MagicMock()
        mock_response.input_tokens = 5

        mock_client.beta.messages.count_tokens.return_value = mock_response

        # 2. Test Count with Caching
        cnt = AnthropicTokenCounter()

        # First call hits API
        assert cnt.count("Hello") == 5
        mock_client.beta.messages.count_tokens.assert_called_once()

        # Second call should hit cache (no new API call)
        assert cnt.count("Hello") == 5
        assert mock_client.beta.messages.count_tokens.call_count == 1

        # New content hits API
        assert cnt.count("New") == 5
        assert mock_client.beta.messages.count_tokens.call_count == 2



from unittest.mock import MagicMock, patch
from typer.testing import CliRunner
from fabra.cli import app

runner = CliRunner()


def test_context_command_success() -> None:
    """Test standard success path with mocked urllib."""
    mock_response = MagicMock()
    mock_response.status = 200
    mock_response.read.return_value = b'{"context_id": "123", "items": []}'
    mock_response.__enter__.return_value = mock_response
    mock_response.__exit__.return_value = None

    with patch("urllib.request.urlopen", return_value=mock_response) as mock_urlopen:
        result = runner.invoke(app, ["context", "explain", "ctx_123"])

        assert result.exit_code == 0
        assert "Fetching trace for ctx_123" in result.stdout
        assert "Context Trace: ctx_123" in result.stdout

        # Verify URL construction
        args, _ = mock_urlopen.call_args
        req = args[0]
        assert req.full_url == "http://127.0.0.1:8000/context/ctx_123/explain"


def test_context_command_server_error() -> None:
    """Test server 500/404 handling."""
    mock_response = MagicMock()
    mock_response.status = 404
    mock_response.__enter__.return_value = mock_response
    mock_response.__exit__.return_value = None

    with patch("urllib.request.urlopen", return_value=mock_response):
        result = runner.invoke(app, ["context", "explain", "ctx_404"])
        assert result.exit_code == 1
        assert "Server returned 404" in result.stdout


def test_context_command_connection_error() -> None:
    """Test connection refused."""
    import urllib.error

    with patch(
        "urllib.request.urlopen",
        side_effect=urllib.error.URLError("Connection refused"),
    ):
        result = runner.invoke(app, ["context", "explain", "ctx_dead"])
        assert result.exit_code == 1
        assert "Connection Failed" in result.stdout



from datetime import datetime, timezone
import re
import os
from typing import List, Optional, Dict, Any
import json
import hashlib
import pandas as pd
from sqlalchemy import text
from sqlalchemy.ext.asyncio import create_async_engine, AsyncEngine
from fabra.store.offline import OfflineStore


class PostgresOfflineStore(OfflineStore):
    def __init__(self, connection_string: str) -> None:
        # Ensure connection string uses asyncpg driver
        if "asyncpg" not in connection_string:
            if "postgresql+psycopg2://" in connection_string:
                connection_string = connection_string.replace(
                    "postgresql+psycopg2://", "postgresql+asyncpg://"
                )
            elif "postgresql://" in connection_string:
                connection_string = connection_string.replace(
                    "postgresql://", "postgresql+asyncpg://"
                )

        # Pool Configuration
        pool_size = int(os.getenv("FABRA_PG_POOL_SIZE", "5"))
        max_overflow = int(os.getenv("FABRA_PG_MAX_OVERFLOW", "10"))

        self.engine: AsyncEngine = create_async_engine(
            connection_string,
            pool_size=pool_size,
            max_overflow=max_overflow,
        )

    async def get_training_data(
        self,
        entity_df: pd.DataFrame,
        features: List[str],
        entity_id_col: str,
        timestamp_col: str = "timestamp",
    ) -> pd.DataFrame:
        # MVP: Similar to DuckDB, we assume features are accessible via SQL.
        # We upload the entity_df to a temporary table and join.

        # Normalize timestamps to UTC-Aware
        # If naive, assume UTC.
        # This ensures consistent behavior with asyncpg/Postgres which prefer Aware datetimes.
        entity_df_norm = entity_df.copy()
        if timestamp_col in entity_df_norm.columns:
            ts_series = entity_df_norm[timestamp_col]
            if (
                pd.api.types.is_datetime64_ns_dtype(ts_series)
                and getattr(ts_series.dtype, "tz", None) is None
            ):
                entity_df_norm[timestamp_col] = ts_series.dt.tz_localize("UTC")
            elif pd.api.types.is_datetime64_dtype(ts_series):
                # Already aware, enable conversion to UTC if mixed?
                # ideally convert to UTC for consistency
                entity_df_norm[timestamp_col] = ts_series.dt.tz_convert("UTC")

        async with self.engine.connect() as conn:  # type: ignore[no-untyped-call]
            # 1. Upload entity_df to temp table
            # Pandas to_sql is sync, so we can't use it directly with async engine easily
            # without running in executor or using a sync connection.
            # For MVP, we'll create the table manually and insert values.
            # This is slow but works for proof of concept.

            # Create temp table
            await conn.execute(
                text(
                    "CREATE TEMP TABLE IF NOT EXISTS temp_entity_lookup (entity_id VARCHAR, timestamp TIMESTAMPTZ)"
                )
            )
            await conn.execute(text("DELETE FROM temp_entity_lookup"))

            # Insert values
            # Insert values
            # Note: We use SQLAlchemy's parameter binding for batch insert (executemany).
            # This is efficient for moderate sizes but slower than Postgres COPY for massive datasets.
            # Production upgrade: Use asyncpg.copy_records_to_table().
            values = []
            for _, row in entity_df_norm.iterrows():
                # Timestamp is now guaranteed Aware UTC by our normalization above
                ts_val = row[timestamp_col].to_pydatetime()
                values.append(
                    {
                        "entity_id": str(row[entity_id_col]),
                        "timestamp": ts_val,
                    }
                )

            if values:
                await conn.execute(
                    text(
                        "INSERT INTO temp_entity_lookup (entity_id, timestamp) VALUES (:entity_id, :timestamp)"
                    ),
                    values,
                )

            await conn.commit()

            # 2. Execute join using LATERAL JOIN for Point-in-Time Correctness
            query_parts = ["SELECT e.*"]
            joins = ""

            import re

            for feature in features:
                if not re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*$", feature):
                    raise ValueError(f"Invalid feature name: {feature}")

                joins += (
                    f" LEFT JOIN LATERAL ("
                    f" SELECT {feature}"
                    f" FROM {feature} f"  # nosec
                    f" WHERE f.entity_id = e.entity_id"
                    f" AND f.timestamp <= e.timestamp"
                    f" ORDER BY f.timestamp DESC"
                    f" LIMIT 1"
                    f" ) {feature}_lat ON TRUE"
                )
                query_parts.append(f", {feature}_lat.{feature} AS {feature}")

            query = "".join(query_parts) + f" FROM temp_entity_lookup e {joins}"

            result = await conn.execute(text(query))  # nosec
            rows = result.fetchall()

            # Convert to DataFrame
            # Convert to DataFrame
            if rows:
                sql_df = pd.DataFrame(rows, columns=result.keys())
            else:
                sql_df = pd.DataFrame(columns=list(result.keys()))

            # Ensure sql_df has UTC timezone to match entity_df_norm
            # asyncpg returns aware datetimes, likely UTC.
            # We assume sql_df is already aware.

            # Merge back to original entity_df (using the normalized one for keys)
            # We merge on entity_id and timestamp.
            # Note: The temp table used 'entity_id' and 'timestamp' as column names.
            # We need to map them back if the original cols were different, but for now we assume standard names
            # or that the caller handles column mapping.
            # Actually, the simplest way is to merge on the index if we preserved order, but SQL doesn't guarantee order.
            # So we merge on the join keys.

            # Ensure join keys match types
            sql_df["entity_id"] = sql_df["entity_id"].astype(str)
            # timestamp might be datetime64[ns] in pandas and datetime in postgres

            # To be safe and simple:
            # 1. Rename sql_df columns to match entity_id_col and timestamp_col if they differ
            if "entity_id" in sql_df.columns and entity_id_col != "entity_id":
                sql_df = sql_df.rename(columns={"entity_id": entity_id_col})
            if "timestamp" in sql_df.columns and timestamp_col != "timestamp":
                sql_df = sql_df.rename(columns={"timestamp": timestamp_col})

            # 2. Merge
            # We use a left join to keep all rows from entity_df_norm
            # We use entity_df_norm for the merge to ensure TZs match (both Aware UTC)
            merged_df_norm = pd.merge(
                entity_df_norm,
                sql_df,
                on=[entity_id_col, timestamp_col],
                how="left",
                suffixes=("", "_sql"),
            )

            # 3. Restore original non-normalized entity_df structure?
            # The caller might expect strict preservation of the input dataframe.
            # But the features are attached. If we return the normalized timestamps (UTC), it's generally cleaner.
            # Let's return the merged result with normalized UTC timestamps.
            return merged_df_norm

    async def execute_sql(self, query: str) -> pd.DataFrame:
        async with self.engine.connect() as conn:  # type: ignore[no-untyped-call]
            result = await conn.execute(text(query))
            rows = result.fetchall()
            if rows:
                return pd.DataFrame(rows, columns=result.keys())
            else:
                return pd.DataFrame(columns=list(result.keys()))

    async def get_historical_features(
        self, entity_name: str, entity_id: str, features: List[str], timestamp: datetime
    ) -> Dict[str, Any]:
        """
        Retrieves feature values as they were at the specified timestamp.
        """
        if not features:
            return {}

        # Use LATERAL JOINs against a single-row virtual table
        selects = []
        joins = ""

        import re

        for feature in features:
            if not re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*$", feature):
                continue

            joins += (
                f" LEFT JOIN LATERAL ("
                f" SELECT {feature}"
                f" FROM {feature} f"  # nosec
                f" WHERE f.entity_id = e.entity_id"
                f" AND f.timestamp <= e.timestamp"
                f" ORDER BY f.timestamp DESC"
                f" LIMIT 1"
                f" ) {feature}_lat ON TRUE"
            )
            selects.append(f"{feature}_lat.{feature} AS {feature}")

        select_clause = ", ".join(selects)

        # Postgres VALUES syntax for virtual table: (VALUES ('id', 'ts'::timestamp)) as e(entity_id, timestamp)
        query = f"""
        SELECT {select_clause}
        FROM (VALUES (:entity_id, :ts::timestamp)) as e(entity_id, timestamp)
        {joins}
        """  # nosec

        async with self.engine.connect() as conn:  # type: ignore[no-untyped-call]
            result = await conn.execute(
                text(query), {"entity_id": str(entity_id), "ts": timestamp}
            )
            row = result.fetchone()
            if row:
                # Convert Row to dict
                return dict(row._mapping)
            return {}

    async def create_index_table(self, index_name: str, dimension: int = 1536) -> None:
        """
        Creates the vector index table if it doesn't exist.
        Schema: id (UUID), entity_id, chunk_index, content, embedding, metadata.
        STRICT SCHEMA: Adds content_hash for deduplication.
        """
        table_name = f"fabra_index_{index_name}"
        async with self.engine.begin() as conn:  # type: ignore[no-untyped-call]
            # Enable extension
            await conn.execute(text("CREATE EXTENSION IF NOT EXISTS vector"))

            # Create table
            await conn.execute(
                text(
                    f"""
                    CREATE TABLE IF NOT EXISTS {table_name} (
                        id UUID DEFAULT gen_random_uuid() PRIMARY KEY,
                        entity_id TEXT NOT NULL,
                        chunk_index INTEGER NOT NULL,
                        content TEXT NOT NULL,
                        content_hash TEXT NOT NULL,
                        embedding vector({dimension}),
                        metadata JSONB DEFAULT '{{}}'::jsonb,
                        created_at TIMESTAMP DEFAULT NOW(),
                        UNIQUE (entity_id, content_hash)
                    )
                    """
                )
            )

            # Create HNSW index
            idx_name = f"idx_{index_name}_embedding"
            await conn.execute(
                text(
                    f"""
                    CREATE INDEX IF NOT EXISTS {idx_name}
                    ON {table_name}
                    USING hnsw (embedding vector_cosine_ops)
                    """
                )
            )

    async def add_documents(
        self,
        index_name: str,
        entity_id: str,
        chunks: List[str],
        embeddings: List[List[float]],
        metadatas: Optional[List[Dict[str, Any]]] = None,
    ) -> None:
        """
        Inserts documents into the index.
        Computes content_hash and adds mandatory metadata.
        """
        if not re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*$", index_name):
            raise ValueError(f"Invalid index name {index_name}. Must be alphanumeric.")

        table_name = f"fabra_index_{index_name}"

        values = []
        for i, (chunk, vec) in enumerate(zip(chunks, embeddings)):
            meta = metadatas[i] if metadatas and i < len(metadatas) else {}

            # Compute hash
            content_hash = hashlib.sha256(chunk.encode("utf-8")).hexdigest()

            # Add Mandatory Metadata
            meta["ingestion_timestamp"] = datetime.now(timezone.utc).isoformat()
            meta["content_hash"] = content_hash
            meta["indexer_version"] = "fabra-v1"

            vec_str = str(vec)

            values.append(
                {
                    "entity_id": entity_id,
                    "chunk_index": i,
                    "content": chunk,
                    "content_hash": content_hash,
                    "embedding": vec_str,
                    "metadata": json.dumps(meta),
                }
            )

        async with self.engine.begin() as conn:  # type: ignore[no-untyped-call]
            # Use ON CONFLICT DO NOTHING to satisfy "prevent duplication" constraint
            insert_query = f"""
                    INSERT INTO {table_name} (entity_id, chunk_index, content, content_hash, embedding, metadata)
                    VALUES (:entity_id, :chunk_index, :content, :content_hash, :embedding, :metadata)
                    ON CONFLICT (entity_id, content_hash) DO NOTHING
                    """  # nosec

            await conn.execute(
                text(insert_query),
                values,
            )

    async def search(
        self,
        index_name: str,
        query_embedding: List[float],
        top_k: int = 5,
        filter_timestamp: Optional[datetime] = None,
    ) -> List[Dict[str, Any]]:
        """
        Performs vector similarity search (Cosine Distance via <=> operator).
        Returns list of dicts with content and metadata.
        """
        if not re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*$", index_name):
            raise ValueError(f"Invalid index name {index_name}. Must be alphanumeric.")

        table_name = f"fabra_index_{index_name}"
        vec_str = str(query_embedding)

        where_clause = ""
        params = {"query_vec": vec_str, "top_k": top_k}

        if filter_timestamp:
            # Filter by ingestion time (created_at)
            where_clause = "WHERE created_at <= :ts"
            params["ts"] = filter_timestamp

        query = f"""
        SELECT content, metadata, 1 - (embedding <=> :query_vec) as score
        FROM {table_name}
        {where_clause}
        ORDER BY embedding <=> :query_vec
        LIMIT :top_k
        """  # nosec

        async with self.engine.connect() as conn:  # type: ignore[no-untyped-call]
            result = await conn.execute(text(query), params)
            rows = result.fetchall()
            return [
                {"content": r.content, "metadata": r.metadata, "score": r.score}
                for r in rows
            ]

    async def _ensure_context_table(self) -> None:
        """Create context_log table if it doesn't exist."""
        async with self.engine.begin() as conn:  # type: ignore[no-untyped-call]
            await conn.execute(
                text(
                    """
                    CREATE TABLE IF NOT EXISTS context_log (
                        context_id VARCHAR(36) PRIMARY KEY,
                        timestamp TIMESTAMPTZ NOT NULL,
                        content TEXT NOT NULL,
                        lineage JSONB NOT NULL,
                        meta JSONB NOT NULL,
                        version VARCHAR(10) DEFAULT 'v1'
                    )
                    """
                )
            )
            # Create index for timestamp-based queries
            await conn.execute(
                text(
                    "CREATE INDEX IF NOT EXISTS idx_context_log_timestamp ON context_log(timestamp)"
                )
            )

    async def log_context(
        self,
        context_id: str,
        timestamp: datetime,
        content: str,
        lineage: Dict[str, Any],
        meta: Dict[str, Any],
        version: str = "v1",
    ) -> None:
        """Persist context to Postgres for replay."""
        await self._ensure_context_table()

        lineage_json = json.dumps(lineage)
        meta_json = json.dumps(meta)

        async with self.engine.begin() as conn:  # type: ignore[no-untyped-call]
            await conn.execute(
                text(
                    """
                    INSERT INTO context_log (context_id, timestamp, content, lineage, meta, version)
                    VALUES (:context_id, :timestamp, :content, :lineage, :meta, :version)
                    ON CONFLICT (context_id) DO UPDATE SET
                        timestamp = EXCLUDED.timestamp,
                        content = EXCLUDED.content,
                        lineage = EXCLUDED.lineage,
                        meta = EXCLUDED.meta,
                        version = EXCLUDED.version
                    """
                ),
                {
                    "context_id": context_id,
                    "timestamp": timestamp,
                    "content": content,
                    "lineage": lineage_json,
                    "meta": meta_json,
                    "version": version,
                },
            )

    async def get_context(self, context_id: str) -> Optional[Dict[str, Any]]:
        """Retrieve a context by ID."""
        await self._ensure_context_table()

        async with self.engine.connect() as conn:  # type: ignore[no-untyped-call]
            result = await conn.execute(
                text("SELECT * FROM context_log WHERE context_id = :context_id"),
                {"context_id": context_id},
            )
            row = result.fetchone()
            if row is None:
                return None

            return {
                "context_id": row.context_id,
                "timestamp": row.timestamp,
                "content": row.content,
                "lineage": row.lineage
                if isinstance(row.lineage, dict)
                else json.loads(row.lineage),
                "meta": row.meta
                if isinstance(row.meta, dict)
                else json.loads(row.meta),
                "version": row.version,
            }

    async def list_contexts(
        self,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        limit: int = 100,
    ) -> List[Dict[str, Any]]:
        """List contexts in time range."""
        await self._ensure_context_table()

        conditions = []
        params: Dict[str, Any] = {"limit": limit}

        if start:
            conditions.append("timestamp >= :start")
            params["start"] = start
        if end:
            conditions.append("timestamp <= :end")
            params["end"] = end

        where_clause = f"WHERE {' AND '.join(conditions)}" if conditions else ""

        query = f"""
            SELECT context_id, timestamp, meta, version
            FROM context_log
            {where_clause}
            ORDER BY timestamp DESC
            LIMIT :limit
        """  # nosec B608 - where_clause built from validated internal conditions

        async with self.engine.connect() as conn:  # type: ignore[no-untyped-call]
            result = await conn.execute(text(query), params)
            rows = result.fetchall()

            results = []
            for row in rows:
                meta = row.meta if isinstance(row.meta, dict) else json.loads(row.meta)
                results.append(
                    {
                        "context_id": row.context_id,
                        "timestamp": row.timestamp,
                        "name": meta.get("name", "unknown"),
                        "token_usage": meta.get("token_usage", 0),
                        "freshness_status": meta.get("freshness_status", "unknown"),
                        "version": row.version,
                    }
                )
            return results



from typer.testing import CliRunner
from fabra.cli import app

runner = CliRunner()


def test_version_command() -> None:
    result = runner.invoke(app, ["version"])
    assert result.exit_code == 0
    assert "Fabra v" in result.stdout


def test_doctor_config_check() -> None:
    """
    Test that 'doctor' runs.
    It might fail or pass depending on the env, but we check it runs
    and outputs something intelligible.
    """
    result = runner.invoke(app, ["doctor"])
    # It might exit with 1 if checks fail (e.g. no redis), so we don't assert 0 blindly
    # unless we mock everything. For now, let's just ensure it calls the command.
    assert "Fabra Doctor" in result.stdout



import pytest
from unittest.mock import AsyncMock
from datetime import datetime
from typing import Dict, Any
from fabra.core import FeatureStore, get_context, get_current_timestamp
from fabra.store.offline import OfflineStore, DuckDBOfflineStore


@pytest.mark.asyncio
async def test_time_travel_context_var() -> None:
    # 1. Verify default is None
    assert get_current_timestamp() is None

    # 2. Verify get_context sets it
    ts = datetime(2023, 1, 1)

    async def mock_func() -> str:
        assert get_current_timestamp() == ts
        return "success"

    res = await get_context(mock_func, timestamp=ts)
    assert res == "success"

    # 3. Verify reset
    assert get_current_timestamp() is None


@pytest.mark.asyncio
async def test_feature_store_time_travel_routing() -> None:
    # Setup
    mock_offline = AsyncMock(spec=OfflineStore)
    mock_online = AsyncMock()

    fs = FeatureStore(offline_store=mock_offline, online_store=mock_online)
    fs.register_entity(name="user", id_column="user_id")
    fs.register_feature(name="f1", entity_name="user", func=lambda x: x)

    mock_offline.get_historical_features.return_value = {"f1": 100}
    mock_online.get_online_features.return_value = {"f1": 200}

    # 1. Real-time call (No timestamp)
    res_rt = await fs.get_online_features("user", "u1", ["f1"])
    # Should call online
    assert res_rt["f1"] == 200
    mock_online.get_online_features.assert_called_once()
    mock_offline.get_historical_features.assert_not_called()

    # Reset mocks
    mock_online.reset_mock()
    mock_offline.reset_mock()

    # 2. Time Travel call
    ts = datetime(2022, 1, 1)

    async def context_logic() -> Dict[str, Any]:
        # Inside context, calling get_online_features
        return await fs.get_online_features("user", "u1", ["f1"])

    res_tt = await get_context(context_logic, timestamp=ts)

    # Should call offline
    assert res_tt["f1"] == 100
    mock_offline.get_historical_features.assert_called_once()
    # Check args
    call_args = mock_offline.get_historical_features.call_args
    assert call_args.kwargs["timestamp"] == ts
    mock_online.get_online_features.assert_not_called()


@pytest.mark.asyncio
async def test_duckdb_historical_query() -> None:
    # Basic smoke test for query construction (using in-memory DuckDB)
    # We need to setup tables first
    store = DuckDBOfflineStore()

    # Create feature table 'f1'
    store.conn.execute(
        "CREATE TABLE f1 (entity_id VARCHAR, timestamp TIMESTAMP, f1 INTEGER)"
    )
    # Insert history
    # u1 at T1=10: value 10
    # u1 at T2=20: value 20
    store.conn.execute("INSERT INTO f1 VALUES ('u1', '2023-01-01 10:00:00', 10)")
    store.conn.execute("INSERT INTO f1 VALUES ('u1', '2023-01-01 12:00:00', 20)")

    # Query at 11:00:00 -> Should get 10
    ts_query = datetime(2023, 1, 1, 11, 0, 0)

    res = await store.get_historical_features("user", "u1", ["f1"], ts_query)
    assert res["f1"] == 10

    # Query at 13:00:00 -> Should get 20
    ts_query_2 = datetime(2023, 1, 1, 13, 0, 0)
    res2 = await store.get_historical_features("user", "u1", ["f1"], ts_query_2)
    assert res2["f1"] == 20



"""Custom exceptions for Fabra."""

from typing import List, Dict, Any


class FabraError(Exception):
    """Base exception for all Fabra errors."""

    pass


class FreshnessSLAError(FabraError):
    """
    Raised when context assembly violates a freshness SLA in strict mode.

    This exception is raised when `freshness_strict=True` is set on a
    `@context` decorator and one or more features exceed the specified
    `freshness_sla` threshold.

    Attributes:
        message: Human-readable error message
        violations: List of features that violated the SLA with details
    """

    def __init__(self, message: str, violations: List[Dict[str, Any]]) -> None:
        super().__init__(message)
        self.message = message
        self.violations = violations

    def __str__(self) -> str:
        violation_details = ", ".join(
            f"{v['feature']} ({v['age_ms']}ms > {v['sla_ms']}ms)"
            for v in self.violations
        )
        return f"{self.message}: {violation_details}"

    def __repr__(self) -> str:
        return f"FreshnessSLAError(message={self.message!r}, violations={self.violations!r})"


class ContextBudgetError(FabraError):
    """Raised when context exceeds max_tokens even after dropping optional items."""

    pass


class StoreConnectionError(FabraError):
    """Raised when unable to connect to a backend store."""

    pass


class FeatureNotFoundError(FabraError):
    """Raised when a requested feature does not exist."""

    pass



import pytest
import time
from fakeredis import FakeStrictRedis
from fabra.scheduler_dist import DistributedScheduler


@pytest.fixture
def redis_client() -> FakeStrictRedis:
    return FakeStrictRedis()


@pytest.fixture
def callback_log() -> list[str]:
    return []


def test_scheduler_starts(redis_client: FakeStrictRedis) -> None:
    scheduler = DistributedScheduler(redis_client)
    scheduler.start()
    try:
        assert scheduler.scheduler.running
    finally:
        scheduler.shutdown()


def test_schedule_job_locks(redis_client: FakeStrictRedis) -> None:
    scheduler = DistributedScheduler(redis_client)
    scheduler.start()

    try:
        # Shared state to verify job execution
        execution_count = {"count": 0}

        def my_job() -> None:
            execution_count["count"] += 1

        # Schedule job with ID "job1"
        scheduler.schedule_job(my_job, interval_seconds=1, job_id="job1")

        # Let it run once
        time.sleep(1.2)

        # Verify logic manually or trust "lock" key presence if we could pause time.
        # Check if lock key exists (it might have expired if job finished fast)
        # But we mostly care that this test doesn't crash or leak.
    finally:
        scheduler.shutdown()


def test_lock_contention(redis_client: FakeStrictRedis) -> None:
    # Simulate two schedulers
    s1 = DistributedScheduler(redis_client)
    s2 = DistributedScheduler(redis_client)
    s1.start()
    s2.start()

    try:
        run_log = []

        def job_logic() -> None:
            run_log.append("run")

        # Manually hold the lock to simulate s1 holding it
        redis_client.set("lock:job_shared", "locked", ex=10)

        # Now schedule on s2
        s2.schedule_job(job_logic, interval_seconds=1, job_id="job_shared")

        # Trigger the job on s2 'manually' via getting the job function from apscheduler
        # This avoids waiting for real time.
        job = s2.scheduler.get_job("job_shared")
        # The 'func' of the job is the 'locked_job' wrapper
        wrapper = job.func

        # Run the wrapper
        wrapper()

        # Since lock is held by "s1" (manually set), s2 should NOT run the logic
        assert len(run_log) == 0

        # Now release lock
        redis_client.delete("lock:job_shared")

        # Run wrapper again
        wrapper()

        # Now it should run
        assert len(run_log) == 1

    finally:
        s1.shutdown()
        s2.shutdown()



import pytest
from unittest.mock import AsyncMock, MagicMock, patch
from fabra.worker import AxiomWorker
from fabra.server import create_app
from fabra.core import FeatureStore
from fastapi.testclient import TestClient


# --- DLQ Tests ---
@pytest.mark.asyncio
async def test_worker_dlq_logic() -> None:
    worker = AxiomWorker(redis_url="redis://mock")
    worker.redis = AsyncMock()

    # Mock message that causes error
    msg_id = "1-0"
    fields = {"data": "invalid-json"}

    # Patch ack method
    with patch.object(worker, "ack", new_callable=AsyncMock) as mock_ack:
        # process_message should catch error and route to DLQ
        await worker.process_message("stream1", msg_id, fields)

        # Verify XADD to DLQ
        assert worker.redis.xadd.called
        args = worker.redis.xadd.call_args
        assert args[0][0] == "fabra:dlq:stream1"
        assert args[0][1]["error"] is not None
        assert args[0][1]["data"] == "invalid-json"

        # Verify ACK
        mock_ack.assert_called_with("stream1", msg_id)


# --- Server Tests ---
@pytest.fixture
def mock_store() -> MagicMock:
    store = MagicMock(spec=FeatureStore)
    store.online_store = MagicMock()
    store.online_store.client = AsyncMock()  # Mock redis client
    # mocked get_online_features needs to return a dict, not a mock
    store.get_online_features = AsyncMock(return_value={"f": 1})
    return store


def test_audit_middleware(mock_store: MagicMock) -> None:
    app = create_app(mock_store)
    client = TestClient(app)

    with patch("fabra.server.logger") as mock_logger:
        # POST request should trigger audit
        client.post(
            "/v1/features",
            json={"entity_name": "E", "entity_id": "1", "features": ["f"]},
        )

        # Verify audit log
        audit_call = [
            c for c in mock_logger.info.call_args_list if "audit_log" in str(c)
        ]
        assert audit_call
        kwargs = audit_call[0].kwargs
        assert kwargs["audit"] is True
        assert kwargs["method"] == "POST"

        # GET request (health) should NOT trigger audit
        mock_logger.reset_mock()
        client.get("/health")
        audit_call_get = [
            c for c in mock_logger.info.call_args_list if "audit_log" in str(c)
        ]
        assert not audit_call_get


@pytest.mark.asyncio
async def test_cache_invalidation(mock_store: MagicMock) -> None:
    app = create_app(mock_store)
    client = TestClient(app)

    # Mock delete
    # Mock delete directly on the online_store, not client
    # because server logic checks hasattr(store.online_store, "delete")
    mock_store.online_store.delete = AsyncMock()

    res = client.delete("/v1/cache/User/u1")
    assert res.status_code == 200
    assert res.json()["status"] == "invalidated"

    mock_store.online_store.delete.assert_called_with("User:u1")



import pytest
from fastapi.testclient import TestClient
from fabra.server import create_app
from fabra.core import FeatureStore
from unittest.mock import MagicMock, AsyncMock


@pytest.fixture
def client() -> TestClient:
    store = MagicMock(spec=FeatureStore)
    store.online_store = MagicMock()
    # Mock get_online_features to return dummy data
    store.get_online_features = AsyncMock(return_value={"age": 30})

    app = create_app(store)
    return TestClient(app)


def test_health_check(client: TestClient) -> None:
    response = client.get("/health")
    assert response.status_code == 200
    assert response.json() == {"status": "ok"}


def test_v1_features_batch(client: TestClient) -> None:
    payload = {"name": "User", "ids": ["u1", "u2"], "features": ["age"]}
    # No API key needed in dev mode if env var not set,
    # but get_api_key logic in server.py defaults to "dev-mode" if not configured.

    response = client.post("/v1/features/batch", json=payload)
    assert response.status_code == 200
    data = response.json()
    assert "u1" in data
    assert "u2" in data
    assert data["u1"]["age"] == 30


def test_v1_features_single(client: TestClient) -> None:
    payload = {"entity_name": "User", "entity_id": "u1", "features": ["age"]}
    response = client.post("/v1/features", json=payload)
    assert response.status_code == 200
    assert response.json() == {"age": 30}
